import importlib
import logging
import time
import enum
from typing import TYPE_CHECKING
from collections.abc import Callable
import ctypes

import numpy as np

from qmi.core.context import QMI_Context
from qmi.core.instrument import QMI_Instrument
from qmi.core.exceptions import QMI_InstrumentException, QMI_TimeoutException, QMI_UnknownNameException
from qmi.core.rpc import rpc_method

if TYPE_CHECKING:
    from picosdk.ps3000a import ps3000a
    from picosdk.ps4000a import ps4000a
    _ps: ps3000a | ps4000a | None = None

else:
    _ps = None

# Global dictionary for the imported library commands.
COMMAND_DICT: dict[str, Callable] = dict()
# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


def _import_modules(library: str) -> None:
    """Import the "picosdk" module. And update the dictionary of commands in the library.

    This import is done in a function, instead of at the top-level, to avoid an unnecessary dependency for
    programs that do not access the instrument directly. NOTE that only ONE module can be loaded at any time.
    If "ps3000a" is imported to _ps, you cannot use Picoscopes based on "ps4000a" libraries with the same instance.

    parameters:
        library: module library string to import. Like "3000a"
    """
    global _ps, COMMAND_DICT
    library = f"ps{library}"
    if _ps is None:
        ps_lib = importlib.import_module(f"picosdk.{library}")
        _ps = getattr(ps_lib, f"{library}")

    # Map generic commands names without model number into the current model's library module commands
    lib_name = _ps.name
    if lib_name != library:
        raise QMI_UnknownNameException(f"Module library {library} does not match loaded {lib_name} module!")

    commands = [_ for _ in _ps.__dict__ if _.startswith("ps")]
    for command in commands:
        COMMAND_DICT[command.lstrip(lib_name)] = getattr(_ps, command)


def _check_error(status: int) -> None:
    """Raise QMI_InstrumentException when the status code is not equal to `PICO_OK`."""
    assert _ps is not None
    if status != _ps.PICO_STATUS["PICO_OK"]:
        raise QMI_InstrumentException("PicoSDK returned status {}".format(_ps.PICO_STATUS_LOOKUP.get(status, "???")))


class ChannelCoupling(enum.IntEnum):
    """Input coupling configurable for each oscilloscope channel."""
    AC = 0
    DC = 1


class TriggerEdge(enum.IntEnum):
    """Configurable trigger edge modes."""
    RISING = 2
    FALLING = 3
    RISING_OR_FALLING = 4
    NONE = 10


class PicoTech_PicoScope(QMI_Instrument):
    """Base class for PicoTech's PicoScope USB oscilloscope instrument drivers.

    Attributes:
        MIN_SAMPLE_VALUE: Min sample value of int16 range.
        MAX_SAMPLE_VALUE: Max sample value of int16 range.
        NUM_CHANNELS:     Number of oscilloscope channels.
        NUM_INPUT_RANGES: Number of supported input ranges. Range '0' is not supported.
    """
    MIN_SAMPLE_VALUE = -32767
    MAX_SAMPLE_VALUE = 32767
    NUM_CHANNELS = 0
    NUM_INPUT_RANGES = 0

    @staticmethod
    def list_instruments(library: str) -> list[str]:
        """Return a list of serial numbers of detected PicoScope instruments of a specific library type.

        Parameters:
            library: The library type to list, e.g. "3000a"

        Returns:
            serials: A list of serial numbers of devices found which use input library type.
        """
        _import_modules(library)
        max_len = 1024
        par_count = ctypes.c_int16(0)
        par_serials = ctypes.create_string_buffer(max_len)
        par_serialslen = ctypes.c_int16(max_len)
        err = COMMAND_DICT["EnumerateUnits"](ctypes.byref(par_count), par_serials, ctypes.byref(par_serialslen))
        _check_error(err)
        serials = par_serials.value.decode("iso8859-1")
        if serials:
            return serials.split(",")
        else:
            return []

    def __init__(self, context: QMI_Context, name: str, serial_number: str) -> None:
        """Initialize the instrument driver.

        Parameters:
            name:          Name for this instrument instance.
            serial_number: Serial number of the Picoscope.
        """
        super().__init__(context, name)
        self._serial_number = serial_number
        self._handle: ctypes.c_int16 | None = None
        self._num_samples: int = 0
        self._time_base_interval_ns: float = 0.0
        self._library: str = ""  # The sw library name, e.g. 3000a

    @property
    def _ps_attr(self):
        assert _ps is not None
        return _ps

    @rpc_method
    def get_input_ranges(self) -> dict[int, float]:
        """Return a dictionary mapping supported input range indexes to the corresponding input range in Volt."""
        self._check_is_open()
        return dict((sel, volt) for (sel, volt) in self._ps_attr.PICO_VOLTAGE_RANGE.items()
                    if sel < self.NUM_INPUT_RANGES)

    @rpc_method
    def open(self) -> None:
        _import_modules(self._library)

        self._check_is_closed()
        _logger.info("[%s] Opening connection to instrument", self._name)

        par_handle = ctypes.c_int16(0)
        par_serial = ctypes.create_string_buffer(self._serial_number.encode("iso8859-1"))
        err = COMMAND_DICT["OpenUnit"](ctypes.byref(par_handle), par_serial)
        _check_error(err)

        self._handle = par_handle

        super().open()

    @rpc_method
    def close(self) -> None:
        self._check_is_open()
        _logger.info("[%s] Closing connection to instrument", self._name)
        COMMAND_DICT["Stop"](self._handle)
        COMMAND_DICT["CloseUnit"](self._handle)
        # Reset values in `close` for eventual re-open.
        self._handle = None
        self._num_samples = 0
        self._time_base_interval_ns = 0.0

        super().close()

    @rpc_method
    def get_device_variant(self) -> str:
        """Get the type of PicoScope (for example "4824")."""
        self._check_is_open()
        max_len = 1024
        par_string = ctypes.create_string_buffer(max_len)
        par_size = ctypes.c_int16(max_len)
        info = self._ps_attr.PICO_INFO["PICO_VARIANT_INFO"]
        err = COMMAND_DICT["GetUnitInfo"](self._handle, par_string, max_len, ctypes.byref(par_size), info)
        _check_error(err)
        return par_string.value.decode("iso8859-1")

    @rpc_method
    def get_serial_number(self) -> str:
        """Get the batch and serial number of the device."""
        self._check_is_open()
        max_len = 1024
        par_string = ctypes.create_string_buffer(max_len)
        par_size = ctypes.c_int16(max_len)
        info = self._ps_attr.PICO_INFO["PICO_BATCH_AND_SERIAL"]
        err = COMMAND_DICT["GetUnitInfo"](self._handle, par_string, max_len, ctypes.byref(par_size), info)
        _check_error(err)
        return par_string.value.decode("iso8859-1")

    @rpc_method
    def set_channel(
        self, channel: int, enable: bool, coupling: ChannelCoupling, range_sel: int, offset: float
            ) -> None:
        """Configure an oscilloscope input channel.

        Parameters:
            channel:    Input channel (range 0 .. 3).
            enable:     True to enable the channel, False to disable it.
            coupling:   Input coupling (`ChannelCoupling.AC` or `ChannelCoupling.DC`).
            range_sel:  Input range selector (range 0 .. 9). See `get_input_ranges()`.
            offset:     Analog input offset in Volt.
                        This offset is added to the analog input signal before digitizing the signal.
        """
        self._check_is_open()
        if (channel < 0) or (channel >= self.NUM_CHANNELS):
            raise ValueError("Invalid channel index")
        if (range_sel < 0) or (range_sel >= self.NUM_INPUT_RANGES):
            raise ValueError("Invalid input range")

        coupling = self._ps_attr.PICO_COUPLING[coupling.name]
        err = COMMAND_DICT["SetChannel"](self._handle, channel, int(enable), coupling, range_sel, offset)
        _check_error(err)

    @rpc_method
    def set_trigger(self, enable: bool, channel: int, threshold: int, edge: TriggerEdge) -> None:
        """Configure the trigger mode of the oscilloscope.

        Parameters:
            enable:     True to enable triggering, False to disable the trigger.
            channel:    Source channel to trigger on (range 0 .. 7).
            threshold:  Trigger level as an ADC code (range -32767 .. +32767).
            edge:       Trigger edge (`TriggerEdge.RISING` or `TriggerEdge.FALLING`
                        or `TriggerEdge.RISING_OR_FALLING`).
        """
        self._check_is_open()
        if (channel < 0) or (channel >= self.NUM_CHANNELS):
            raise ValueError("Invalid channel index")
        if (threshold < self.MIN_SAMPLE_VALUE) or (threshold > self.MAX_SAMPLE_VALUE):
            raise ValueError("Invalid threshold")

        err = COMMAND_DICT["SetSimpleTrigger"](self._handle, int(enable), channel, threshold, int(edge), 0, 0)
        _check_error(err)

    @rpc_method
    def disable_trigger(self):
        """ Disables all triggers in all channels """
        self._check_is_open()
        err = COMMAND_DICT["SetTriggerChannelConditions"](self._handle, None, 0)
        _check_error(err)

    @rpc_method
    def stop(self) -> None:
        """Stop any ongoing acquisition."""
        self._check_is_open()
        err = COMMAND_DICT["Stop"](self._handle)
        _check_error(err)

    @rpc_method
    def is_block_ready(self) -> bool:
        """Return True if a block acquisition has completed, or False if the acquisition is still in progress.
        Only call this function after starting an acquisition via `run_block()`.
        """
        self._check_is_open()
        par_ready = ctypes.c_int16()
        err = COMMAND_DICT["IsReady"](self._handle, ctypes.byref(par_ready))
        _check_error(err)
        return par_ready.value != 0

    @rpc_method
    def wait_block_ready(self, timeout: float) -> None:
        """Wait until the running block acquisition has completed. Uses the "is_block_ready" above.

        Parameters:
            timeout: Maximum time to wait in seconds.

        Raises:
            ValueError:           If the timeout input value is negative.
            QMI_TimeoutException: If the timeout expires before a block is ready.
        """
        self._check_is_open()
        if timeout < 0:
            raise ValueError("Invalid timeout")

        endtime = time.monotonic() + timeout
        while True:
            par_ready = self.is_block_ready()
            if par_ready:
                break
            elif time.monotonic() >= endtime:
                raise QMI_TimeoutException("Timeout while waiting for block from PicoScope")

            time.sleep(0.01)

    @rpc_method
    def get_block_data(self, channels: list[int]) -> tuple[np.ndarray, float, list[bool]]:
        """Retrieve a block of data, previously acquired via `run_block()`.

        Samples are returned as 16-bit signed integers, where the value
        (2**16-1) corresponds to the maximum positive voltage and -(2**16-1)
        corresponds to the most negative voltage in the input range.

        The number of retrieved samples corresponds to the sum of pre-trigger
        and post-trigger samples as specified in the call to `run_block()`.

        Parameters:
            channels: List of channels for which to retrieve samples and overrange indications.

        Returns:
            samples:   A 2D Numpy array with shape (num_channels, num_samples) containing signed 16-bit samples.
            time_base_interval_ns: The time interval between samples as a floating point number in nanoseconds.
            overrange: A list of booleans indicating whether overvoltage occurred on the selected channels.

        Raises:
            ValueError:              If a channel number in 'channels' is invalid.
            QMI_InstrumentException: If unexpected number of samples were obtained.
        """
        self._check_is_open()

        for chan in channels:
            if (chan < 0) or (chan >= self.NUM_CHANNELS):
                raise ValueError("Invalid channel index")

        # Allocate data buffers.
        buffers: list[ctypes.Array[ctypes.c_int16]] = []
        for chan in range(self.NUM_CHANNELS):

            # Allocate real buffer for selected channels, zero-length buffer for non-selected channels.
            if chan in channels:
                buf_len = self._num_samples
            else:
                buf_len = 0
            buf = (ctypes.c_int16 * buf_len)()
            buffers.append(buf)

            # Register allocated buffer with driver.
            err = COMMAND_DICT["SetDataBuffer"](self._handle, chan, buf, buf_len, 0, 0)
            _check_error(err)

        # Retrieve data.
        par_num_samples = ctypes.c_uint32(self._num_samples)
        par_overflow = ctypes.c_int16(0)
        err = COMMAND_DICT["GetValues"](
            self._handle,
            0,  # startIndex
            ctypes.byref(par_num_samples),
            1,  # downSampleRatio
            0,  # downSampleRatioMode
            0,  # segmentIndex
            ctypes.byref(par_overflow),
        )
        _check_error(err)

        got_samples = par_num_samples.value
        if got_samples != self._num_samples:
            raise QMI_InstrumentException(
                "Got {} samples from PicoScope while expecting {} samples".format(got_samples, self._num_samples)
            )

        # Convert data to Numpy array.
        samples: np.typing.NDArray[np.int16] = np.empty((len(channels), self._num_samples), dtype=np.int16)
        for (i, chan) in enumerate(channels):
            samples[i] = np.ctypeslib.as_array(buffers[chan])

        # Extract overrange markers.
        overrange = [bool((par_overflow.value >> chan) & 1) for chan in channels]

        return samples, self._time_base_interval_ns, overrange

    @rpc_method
    def run_block(self, num_pretrig_samples: int, num_posttrig_samples: int, time_base: int) -> None:
        raise NotImplementedError("Method 'run_block' is not implemented in the base class")

    @rpc_method
    def get_sampling_interval(self, time_base: int) -> float:
        raise NotImplementedError("Method 'get_sampling_interval' is not implemented in the base class")



"""Instrument driver for the AimTTi TGF3000 Series and TGF4000 Series Function Generator / Frequency Counter."""

import enum
import logging
import math
import time
from typing import NamedTuple

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.transport import create_transport


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class FrequencyMeasurement(NamedTuple):
    """Frequency measurement produced by the instrument.

    Attributes:
        timestamp: Approximate POSIX timestamp of measurement.
        frequency: Measured frequency in Hz.
    """
    timestamp: float
    frequency: float


class WaveformType(enum.Enum):
    """Waveform types."""
    SINE = "SINE"
    SQUARE = "SQUARE"
    RAMP = "RAMP"
    NOISE = "NOISE"


class CounterInputChannel(enum.Enum):
    """Counter input channels."""
    TRIG_IN = "DC"
    REF_IN = "AC"


class TT_TGF_3000_4000_Series(QMI_Instrument):
    """Instrument driver for the TGF3000 and TGF4000 Series function generators.

    This driver has been tested with TGF3162 and TGF4242.
    It probably works with all TGF3000 and TGF4000 Series instruments.

    This driver supports only a subset of the functionality of the instrument.
    """

    # Instrument should respond within 4 seconds.
    RESPONSE_TIMEOUT = 4.0

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        """Initialize driver.

        To access the instrument via Ethernet, provide a transport descriptor like::
          "tcp:172.16.0.100".

        This driver does not support connecting to the instrument via USB.

        Parameters:
            context:    QMI_Context instance.
            name:       Name for this instrument instance.
            transport:  QMI transport descriptor for the SCPI channel.
        """
        super().__init__(context, name)

        self._scpi_transport = create_transport(transport, default_attributes={"port": 9221})
        self._scpi_protocol = ScpiProtocol(self._scpi_transport,
                                           command_terminator="\n",
                                           response_terminator="\r\n",
                                           default_timeout=self.RESPONSE_TIMEOUT)

    @rpc_method
    def open(self) -> None:
        _logger.info("[%s] Opening connection to instrument", self._name)
        self._scpi_transport.open()
        try:
            # Clear error queue to avoid reporting stale errors.
            self._scpi_protocol.write("*CLS")
        except OSError:
            self._scpi_transport.close()
            raise
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("[%s] Closing connection to instrument", self._name)
        super().close()
        self._scpi_transport.close()

    def _write(self, cmd: str) -> None:
        """Send a SCPI command."""
        self._scpi_protocol.write(cmd)

    def _ask(self, cmd: str) -> str:
        """Send an SCPI command, then read and return the response."""
        return self._scpi_protocol.ask(cmd)

    @staticmethod
    def _check_channel(channel: int) -> None:
        """Check that the channel number is valid."""
        if channel not in (1, 2):
            raise ValueError(f"Invalid channel number {channel!r}")

    def _check_error(self) -> None:
        """Read the instrument error queue and raise an exception if there is an error."""

        # Read error queue.
        resp = self._ask("EER?")

        # Extract error code; 0 means no error occurred.
        try:
            error_code = int(resp)
        except ValueError:
            # pylint: disable=raise-missing-from
            raise QMI_InstrumentException(f"Unexpected response to 'EER?': {resp!r}")

        if error_code != 0:
            # Report error.
            raise QMI_InstrumentException(f"Instrument returned error code {error_code}")

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance."""
        resp = self._ask("*IDN?")
        words = resp.rstrip().split(",")
        if len(words) != 4:
            raise QMI_InstrumentException(f"Unexpected response to *IDN?, got {resp!r}")
        return QMI_InstrumentIdentification(vendor=words[0].strip(),
                                            model=words[1].strip(),
                                            serial=words[2].strip(),
                                            version=words[3].strip())

    @rpc_method
    def reset(self) -> None:
        """Reset the instrument, returning most settings to defaults."""
        self._write("*CLS")
        self._write("*RST")
        self._check_error()

    @rpc_method
    def set_output_enabled(self, channel: int, enable: bool) -> None:
        """Enable or disable the specified output channel.

        Parameters:
            channel: Channel index (1 or 2).
            enable:  True to enable output, False to disable.
        """
        self._check_channel(channel)
        arg = "ON" if enable else "OFF"
        self._write(f"CHN {channel};OUTPUT {arg}")
        self._check_error()

    @rpc_method
    def set_output_inverted(self, channel: int, invert: bool) -> None:
        """Enable or disable inverted output.

        Parameters:
            channel: Channel index (1 or 2).
            invert:  True to enable inverted output; False to disable.
        """
        self._check_channel(channel)
        arg = "INVERT" if invert else "NORMAL"
        self._write(f"CHN {channel};OUTPUT {arg}")
        self._check_error()

    @rpc_method
    def set_output_load_impedance(self, channel: int, impedance: float) -> None:
        """Configure the assumed load impedance.

        Note: The actual output impedance of the instrument is fixed at 50 Ohm.
        This command configures the assumed impedance of the load that is
        connected to the output. This affects the calculation of output voltage levels.

        Parameters:
            channel:    Channel index (1 or 2).
            impedance:  Assumed load impedance in Ohm, range 1 to 10000, or `math.inf` to assume open circuit.
        """
        self._check_channel(channel)
        arg = "OPEN" if math.isinf(impedance) else f"{impedance:.0f}"
        self._write(f"CHN {channel};ZLOAD {arg}")
        self._check_error()

    @rpc_method
    def set_waveform(self, channel: int, waveform: WaveformType) -> None:
        """Set waveform type.

        This driver currently supports only a subset of the waveforms

        This command may fail if the selected waveform does not support
        the current frequency. In that case, the frequency must first
        be adjusted by calling `set_frequency()`.

        Parameters:
            channel:  Channel index (1 or 2).
            waveform: Waveform type (a `WaveformType.XXX` constant).
        """
        self._check_channel(channel)
        arg = WaveformType(waveform).value
        self._write(f"CHN {channel};WAVE {arg}")
        self._check_error()

    @rpc_method
    def set_frequency(self, channel: int, frequency: float) -> None:
        """Change the output frequency.

        This function is only supported for waveform types SINE, SQUARE and RAMP.

        Parameters:
            channel:   Channel index (1 or 2).
            frequency: Frequency in Hz.
        """
        self._check_channel(channel)
        self._write(f"CHN {channel};FREQ {frequency:.6f}")
        self._check_error()

    @rpc_method
    def set_square_duty_cycle(self, channel: int, duty_cycle: float) -> None:
        """Set the duty cycle of the square wave.

        This function is only supported for waveform type SQUARE.

        Parameters:
            channel:    Channel index (1 or 2).
            duty_cycle: Duty cycle as a percentage between  0.0 and 100.0;
                        actual range also depends on the frequency.
        """
        self._check_channel(channel)
        self._write(f"CHN {channel};SQRSYMM {duty_cycle:.3f}")
        self._check_error()

    @rpc_method
    def set_ramp_symmetry(self, channel: int, symmetry: float) -> None:
        """Set the symmetry of the ramp waveform.

        The symmetry represents the duration of the rising edge as a percentage
        of the waveform period.

        Symmetry level 50% means a symmetric triangular wave where rising edge
        and falling edge have the same duration.
        Symmetry level 0% means instantaneous rise followed by slow falling edge.
        Symmetry level 100% means slow rising edge followed by instantaneous drop.

        This function is only supported for waveform type RAMP.

        Parameters:
            channel:   Channel index (1 or 2).
            symmetry:  Symmetry as a percentage in range 0.0 to 100.0.
        """
        self._check_channel(channel)
        self._write(f"CHN {channel};RMPSYMM {symmetry:.3f}")
        self._check_error()

    @rpc_method
    def set_amplitude(self, channel: int, amplitude: float, offset: float) -> None:
        """Set output amplitude and offset.

        Note that both amplitude and offset are configured with respect to
        an assumed output load impedance. If the configured impedance does not
        match the actual load impedance, the actual output voltage will not
        match the configured voltage.

        Parameters:
            channel:   Channel index (1 or 2).
            amplitude: Peak-to-peak signal amplitude in Vpp.
            offset:    Offset level in Volt.
        """
        self._check_channel(channel)
        self._write(f"CHN {channel};AMPL {amplitude:.3f};DCOFFS {offset:.3f}")
        self._check_error()

    @rpc_method
    def set_phase(self, channel: int, phase: float) -> None:
        """Set the phase offset of the waveform.

        Parameters:
            channel:    Channel index (1 or 2).
            phase:      Phase offset in degrees (range 0.0 to 360.0).
        """
        self._check_channel(channel)
        self._write(f"CHN {channel};PHASE {phase:.3f}")
        self._check_error()

    @rpc_method
    def get_external_reference_enabled(self) -> bool:
        """Return True if the external reference clock input is selected."""
        resp = self._ask("CLKSRC?")
        if resp.strip().upper() == "EXT":
            return True
        elif resp.strip().upper() == "INT":
            return False
        raise QMI_InstrumentException(f"Unexpected response to CLKSRC?, got {resp!r}")

    @rpc_method
    def set_external_reference_enabled(self, enable: bool) -> None:
        """Enable or disable external 10 MHz reference clock input.

        Parameters:
            enable: True to enable the external reference clock input;
                    False to select the internal reference clock.
        """
        arg = "EXT" if enable else "INT"
        self._write(f"CLKSRC {arg}")
        self._check_error()

    @rpc_method
    def set_counter_enabled(self, enable: bool) -> None:
        """Enable or disable the frequency counter function.

        Parameters:
            enable: True to enable the frequency counter; False to disable.
        """
        arg = "ON" if enable else "OFF"
        self._write(f"CNTRSWT {arg}")
        self._check_error()

    @rpc_method
    def set_counter_input(self, input: CounterInputChannel) -> None:
        """Select the counter input channel.

        Parameters:
            input: `CounterInputChannel.TRIG_IN` or `CounterInputChannel.REF_IN`.
        """
        arg = CounterInputChannel(input).value
        self._write(f"CNTRCPLNG {arg}")
        self._check_error()

    @rpc_method
    def read_frequency(self) -> FrequencyMeasurement:
        """Read the currently measured counter value.

        This function only supports frequency measurement mode.
        The alternative measurement modes (period, width and duty-cycle) are
        not supported.

        Returns:
            Measured frequency as a `FrequencyMeasurement` tuple.
        """

        resp = self._ask("CNTRVAL?")
        timestamp = time.time()

        if not resp.endswith("Hz"):
            raise QMI_InstrumentException(
                f"Unexpected response to 'CNTRVAL?': {resp!r}. Is counter in frequency mode?")

        try:
            # Strip "Hz" suffix and convert to floating point.
            frequency = float(resp[:-2])
        except ValueError:
            # pylint: disable=raise-missing-from
            raise QMI_InstrumentException(f"Unexpected response to 'CNTRVAL?': {resp!r}")

        return FrequencyMeasurement(timestamp, frequency)

    @rpc_method
    def set_local(self) -> None:
        """Switch the instrument to local mode.

        Every remote command or query causes the instrument to enter remote mode.
        In remote mode, most front-panel keys are not functional.

        This function returns the instrument to `local` mode.
        It is equivalent to pushing the "LOCAL" button on the front panel.
        """
        self._write("LOCAL")


class TT_TGF3162(TT_TGF_3000_4000_Series):
    """Instrument driver for the AimTTi TGF3162 signal generator / frequency counter."""
    pass



from collections import namedtuple
import logging
import unittest
from unittest.mock import Mock, call, patch

import qmi
from qmi.core.exceptions import QMI_TimeoutException, QMI_InstrumentException
from qmi.core.transport import QMI_Transport
from qmi.instruments.nkt_photonics import NktPhotonics_KoherasAdjustik
from qmi.instruments.nkt_photonics.nkt_photonics_interbus_protocol import NKTPhotonicsInterbusProtocol


class TestTransportOperations(unittest.TestCase):

    def setUp(self) -> None:
        # Add patches
        patcher = patch(
            'qmi.instruments.nkt_photonics.adjustik.create_transport', spec=QMI_Transport)
        self.transport = patcher.start().return_value
        self.addCleanup(patcher.stop)
        qmi.start("test_adjustik_open_close")
        self.laser = qmi.make_instrument("test_laser", NktPhotonics_KoherasAdjustik, "transport_stub")

    def tearDown(self) -> None:
        if self.laser.is_open():
            self.laser.close()
        qmi.stop()

    def test_open(self):
        # arrange
        expected_calls = [call.open(), call.discard_read()]
        # act
        self.laser.open()
        # assert
        self.assertEqual(self.transport.mock_calls, expected_calls)

    def test_close(self):
        # arrange
        expected_calls = [call.close]
        # act
        self.laser.open()
        self.transport.reset_mock()
        self.laser.close()
        # assert
        self.assertEqual(self.transport.mock_calls, expected_calls)


class TestMethods(unittest.TestCase):

    def setUp(self) -> None:
        qmi.start("test_adjustik_methods")

        patcher = patch(
            'qmi.instruments.nkt_photonics.adjustik.create_transport', spec=QMI_Transport)
        self.transport = patcher.start().return_value
        self.addCleanup(patcher.stop)
        patcher = patch(
            'qmi.instruments.nkt_photonics.adjustik.NKTPhotonicsInterbusProtocol',
            spec=NKTPhotonicsInterbusProtocol
        )
        self.interbus = patcher.start().return_value
        self.addCleanup(patcher.stop)

        self.laser = qmi.make_instrument("test_laser", NktPhotonics_KoherasAdjustik, "transport_stub")
        self.laser.open()

    def tearDown(self) -> None:
        if self.laser.is_open():
            self.laser.close()
        qmi.stop()

    def test_getters(self):
        TestDescr = namedtuple("TestDescriptor", ["method_name", "register", "value", "value_bytes"])
        test_list = [
            TestDescr("get_basik_module_type",                          0x61,       (0x33), b"\x33"),
            TestDescr("get_basik_emission",                             0x30,          (1), b"\x01"),
            TestDescr("get_basik_setup_bits",                           0x31,     (0xFFFF), b"\xff\xff"),
            TestDescr("get_basik_output_power_setpoint_mW",             0x22,        (3.0), b',\x01'),
            TestDescr("get_basik_output_power_setpoint_dBm",            0xa0,        (3.0), b',\x01'),
            TestDescr("get_basik_wavelength_offset_setpoint",           0x2a,     (1000.0), b"\x10'"),
            TestDescr("get_basik_user_area",                            0x8d, (bytes(240)), bytes(240)),
            TestDescr("get_basik_status_bits",                          0x66,     (0xFFFF), b"\xff\xff"),
            TestDescr("get_basik_output_power_mW",                      0x17,        (3.0), b',\x01'),
            TestDescr("get_basik_output_power_dBm",                     0x90,        (3.0), b',\x01'),
            TestDescr("get_basik_standard_wavelength",                  0x32,     (6150.5), b'\x41\xf0\x00\x00'),
            TestDescr("get_basik_wavelength_offset",                    0x72,     (6150.5), b'\x41\xf0\x00\x00'),
            TestDescr("get_basik_module_temperature",                   0x1c,       (1000), b"\x10'"),
            TestDescr("get_basik_module_supply_voltage",                0x1e,         (10), b"\x10'"),
            TestDescr("get_basik_module_wavelength_modulation_enabled", 0xb5,          (1), b"\x01"),
            TestDescr("get_basik_wavelength_modulation_frequency",      0xb8,    (2.0, 1.0), b'\x00\x00\x00\x40\x00\x00\x80\x3f'),
            TestDescr("get_basik_wavelength_modulation_level",          0x2b,       (1000), b"\x10'"),
            TestDescr("get_basik_wavelength_modulation_offset",         0x2f,       (1000), b"\x10'"),
            TestDescr("get_basik_amplitude_modulation_frequency",       0xba,    (2.0, 1.0), b'\x00\x00\x00\x40\x00\x00\x80\x3f'),
            TestDescr("get_basik_amplitude_modulation_depth",           0x2c,       (1000), b"\x10'"),
            TestDescr("get_basik_modulation_setup_bits",                0xb7,     (0xFFFF), b"\xff\xff"),
            TestDescr("get_adjustik_module_type",                       0x61,       (0x34), b"\x34"),
        ]
        for test in test_list:
            self.interbus.reset_mock()
            with self.subTest(i=test.method_name):
                # arrange per test
                self.interbus.get_register.return_value = test.value_bytes
                expected_dest = 0x01 if "basik" in test.method_name else 0x80 if "adjustik" in test.method_name else 0x00
                # act
                actual_rv = getattr(self.laser, test.method_name)()
                # assert
                self.assertEqual(actual_rv, test.value)
                self.interbus.get_register.assert_called_once_with(expected_dest, test.register)

    def test_setters(self):
        TestDescr = namedtuple("TestDescriptor", ["method_name", "register", "value", "value_bytes"])
        test_list = [
            TestDescr("set_basik_setup_bits",                      0x31,  (0xFFFF,), b"\xff\xff"),
            TestDescr("set_basik_output_power_setpoint_mW",        0x22,     (3.0,), b',\x01'),
            TestDescr("set_basik_output_power_setpoint_dBm",       0xa0,     (3.0,), b',\x01'),
            TestDescr("set_basik_wavelength_offset_setpoint",      0x2a,  (1000.0,), b"\x10'"),
            TestDescr("set_basik_wavelength_modulation_frequency", 0xb8,  (2.0, 1.0), b'\x00\x00\x00\x40\x00\x00\x80\x3f'),
            TestDescr("set_basik_wavelength_modulation_level",     0x2b,  (1000.0,), b"\x10'"),
            TestDescr("set_basik_wavelength_modulation_offset",    0x2f,  (1000.0,), b"\x10'"),
            TestDescr("set_basik_modulation_setup_bits",           0xb7,  (0xFFFF,), b"\xff\xff"),
        ]
        for test in test_list:
            self.interbus.reset_mock()
            with self.subTest(i=test.method_name):
                # arrange
                expected_dest = 0x01 if "basik" in test.method_name else 0x80 if "adjustik" in test.method_name else 0x00
                # act
                getattr(self.laser, test.method_name)(*test.value)
                # assert
                self.interbus.set_register.assert_called_once_with(expected_dest, test.register, test.value_bytes)

    def test_enable_basik_emission_enables_emission(self):
        self.interbus.reset_mock()
        # arrange
        expected_dest = 0x01
        # act
        getattr(self.laser, "enable_basik_emission")()
        # assert
        self.interbus.set_register.assert_called_once_with(expected_dest, 0x30, b"\x01")

    def test_disable_basik_emission_disables_emission(self):
        self.interbus.reset_mock()
        # arrange
        expected_dest = 0x01
        # act
        getattr(self.laser, "disable_basik_emission")()
        # assert
        self.interbus.set_register.assert_called_once_with(expected_dest, 0x30, b"\x00")


class TestInterbus(unittest.TestCase):

    ADJUSTIK_TIMEOUT = 1.0

    def setUp(self):

        # Suppress logging.
        logging.getLogger("qmi.instruments.nkt_photonics.nkt_photonics_interbus_protocol").setLevel(logging.CRITICAL)

        context = Mock()
        self._mock_transport = Mock()

        with patch("qmi.instruments.nkt_photonics.adjustik.create_transport", return_value=self._mock_transport):
            self._laser = NktPhotonics_KoherasAdjustik(context, "laser", "transport_stub")

        self._laser.open()

        self.assertEqual(self._mock_transport.mock_calls, [
            call.open(),
            call.discard_read()
        ])
        self._mock_transport.reset_mock()

    def tearDown(self):
        logging.getLogger("qmi.instruments.nkt_photonics.nkt_photonics_interbus_protocol").setLevel(logging.NOTSET)

    def test_get_request(self):

        # Request: dst=1, src=161, MessageType.READ, reg=0x32
        request_msg = bytes([0x0d, 0x01, 0xa1, 0x04, 0x32, 0x26, 0xcd, 0x0a])

        # Response: dst=161, src=1, MessageType.DATAGRAM, reg=0x32, data=uint32(10644000)
        response_msg = bytes([0x0d, 0xa1, 0x01, 0x08, 0x32, 0x20, 0x6a, 0xa2, 0x00, 0x2c, 0x8c, 0x0a])

        self._mock_transport.read_until.return_value = response_msg

        v = self._laser.get_basik_standard_wavelength()
        self.assertAlmostEqual(v, 1064400.0)

        self.assertEqual(self._mock_transport.mock_calls, [
            call.write(request_msg),
            call.read_until(message_terminator=b"\n", timeout=self.ADJUSTIK_TIMEOUT)
        ])

    def test_set_request(self):

        # Request: dst=1, src=161, MessageType.WRITE, reg=0x2a, data=int16(840)
        request_msg = bytes([0x0d, 0x01, 0xa1, 0x05, 0x2a, 0x48, 0x03, 0x8c, 0xd1, 0x0a])

        # Response: dst=161, src=1, MessageType.ACK, reg=0x2a
        response_msg = bytes([0x0d, 0xa1, 0x01, 0x03, 0x2a, 0x7b, 0x89, 0x0a])

        self._mock_transport.read_until.return_value = response_msg

        self._laser.set_basik_wavelength_offset_setpoint(84.0)

        self.assertEqual(self._mock_transport.mock_calls, [
            call.write(request_msg),
            call.read_until(message_terminator=b"\n", timeout=self.ADJUSTIK_TIMEOUT)
        ])

    def test_two_requests(self):

        # Request: dst=1, src=161, MessageType.READ, reg=0x32
        request_msg_1 = bytes([0x0d, 0x01, 0xa1, 0x04, 0x32, 0x26, 0xcd, 0x0a])

        # Response: dst=161, src=1, MessageType.DATAGRAM, reg=0x32, data=uint32(10644000)
        response_msg_1 = bytes([0x0d, 0xa1, 0x01, 0x08, 0x32, 0x20, 0x6a, 0xa2, 0x00, 0x2c, 0x8c, 0x0a])

        # Request: dst=1, src=162, MessageType.READ, reg=0x2a
        request_msg_2 = bytes([0x0d, 0x01, 0xa2, 0x04, 0x2a, 0xec, 0xa4, 0x0a])

        # Response: dst=162, src=1, MessageType.DATAGRAM, reg=0x2a, data=int16(740)
        response_msg_2 = bytes([0x0d, 0xa2, 0x01, 0x08, 0x2a, 0xe4, 0x02, 0xf8, 0xc9, 0x0a])

        self._mock_transport.read_until.side_effect = [response_msg_1, response_msg_2]

        v = self._laser.get_basik_standard_wavelength()
        self.assertAlmostEqual(v, 1064400.0)

        v = self._laser.get_basik_wavelength_offset_setpoint()
        self.assertAlmostEqual(v, 74.0)

        self.assertEqual(self._mock_transport.mock_calls, [
            call.write(request_msg_1),
            call.read_until(message_terminator=b"\n", timeout=self.ADJUSTIK_TIMEOUT),
            call.write(request_msg_2),
            call.read_until(message_terminator=b"\n", timeout=self.ADJUSTIK_TIMEOUT)
        ])

    def test_retry_after_timeout(self):

        # Request: dst=1, src=161, MessageType.READ, reg=0x32
        request_msg = bytes([0x0d, 0x01, 0xa1, 0x04, 0x32, 0x26, 0xcd, 0x0a])

        # Response: dst=161, src=1, MessageType.DATAGRAM, reg=0x32, data=uint32(10644000)
        response_msg = bytes([0x0d, 0xa1, 0x01, 0x08, 0x32, 0x20, 0x6a, 0xa2, 0x00, 0x2c, 0x8c, 0x0a])

        # First time raise QMI_TimeoutException. Second time return expected response.
        read_until_data = [QMI_TimeoutException(), response_msg]

        def fake_read_until(message_terminator, timeout):
            v = read_until_data.pop(0)
            if isinstance(v, Exception):
                raise v
            else:
                return v

        self._mock_transport.read_until.side_effect = fake_read_until

        v = self._laser.get_basik_standard_wavelength()
        self.assertAlmostEqual(v, 1064400.0)

        self.assertEqual(self._mock_transport.mock_calls, [
            call.write(request_msg),
            call.read_until(message_terminator=b"\n", timeout=self.ADJUSTIK_TIMEOUT),
            call.write(request_msg),
            call.read_until(message_terminator=b"\n", timeout=self.ADJUSTIK_TIMEOUT)
        ])

    def test_retry_after_bad_message(self):

        # Request: dst=1, src=161, MessageType.READ, reg=0x32
        request_msg = bytes([0x0d, 0x01, 0xa1, 0x04, 0x32, 0x26, 0xcd, 0x0a])

        # Response: dst=161, src=1, MessageType.DATAGRAM, reg=0x32, data=uint32(10644000)
        response_msg = bytes([0x0d, 0xa1, 0x01, 0x08, 0x32, 0x20, 0x6a, 0xa2, 0x00, 0x2c, 0x8c, 0x0a])

        # First time return garbage. Second time return expected response.
        self._mock_transport.read_until.side_effect = [b"\raap noot\n", response_msg]

        v = self._laser.get_basik_standard_wavelength()
        self.assertAlmostEqual(v, 1064400.0)

        self.assertEqual(self._mock_transport.mock_calls, [
            call.write(request_msg),
            call.read_until(message_terminator=b"\n", timeout=self.ADJUSTIK_TIMEOUT),
            call.write(request_msg),
            call.read_until(message_terminator=b"\n", timeout=self.ADJUSTIK_TIMEOUT)
        ])

    def test_nack_response(self):

        # Request: dst=1, src=161, MessageType.READ, reg=0x32
        request_msg = bytes([0x0d, 0x01, 0xa1, 0x04, 0x32, 0x26, 0xcd, 0x0a])

        # Response: dst=161, src=1, MessageType.NACK, reg=0x32
        response_msg = bytes([0x0d, 0xa1, 0x01, 0x00, 0x32, 0xbd, 0xe3, 0x0a])

        # First time return garbage. Second time return expected response.
        self._mock_transport.read_until.return_value = response_msg

        with self.assertRaises(QMI_InstrumentException):
            self._laser.get_basik_standard_wavelength()

        self.assertEqual(self._mock_transport.mock_calls, [
            call.write(request_msg),
            call.read_until(message_terminator=b"\n", timeout=self.ADJUSTIK_TIMEOUT),
        ])

    def test_unsync_after_timeout(self):

        # Request: dst=1, src=161, MessageType.READ, reg=0x32
        request_msg_1 = bytes([0x0d, 0x01, 0xa1, 0x04, 0x32, 0x26, 0xcd, 0x0a])

        # Response: dst=161, src=1, MessageType.DATAGRAM, reg=0x32, data=uint32(10644000)
        response_msg_1 = bytes([0x0d, 0xa1, 0x01, 0x08, 0x32, 0x20, 0x6a, 0xa2, 0x00, 0x2c, 0x8c, 0x0a])

        # Request: dst=1, src=162, MessageType.READ, reg=0x2a
        request_msg_2 = bytes([0x0d, 0x01, 0xa2, 0x04, 0x2a, 0xec, 0xa4, 0x0a])

        # Response: dst=162, src=1, MessageType.DATAGRAM, reg=0x2a, data=int16(740)
        response_msg_2 = bytes([0x0d, 0xa2, 0x01, 0x08, 0x2a, 0xe4, 0x02, 0xf8, 0xc9, 0x0a])

        self._mock_transport.read_until.side_effect = [response_msg_1, response_msg_2]

        # 1st call: Raise QMI_TimeoutException.
        # 2nd call: Return expected response.
        # 3rd call: Return duplicate of first response.
        # 4th call: Return expected response.
        read_until_data = [QMI_TimeoutException(), response_msg_1, response_msg_1, response_msg_2]

        def fake_read_until(message_terminator, timeout):
            v = read_until_data.pop(0)
            if isinstance(v, Exception):
                raise v
            else:
                return v

        self._mock_transport.read_until.side_effect = fake_read_until

        v = self._laser.get_basik_standard_wavelength()
        self.assertAlmostEqual(v, 1064400.0)

        v = self._laser.get_basik_wavelength_offset_setpoint()
        self.assertAlmostEqual(v, 74.0)

        self.assertEqual(self._mock_transport.mock_calls, [
            call.write(request_msg_1),
            call.read_until(message_terminator=b"\n", timeout=self.ADJUSTIK_TIMEOUT),
            call.write(request_msg_1),
            call.read_until(message_terminator=b"\n", timeout=self.ADJUSTIK_TIMEOUT),
            call.write(request_msg_2),
            call.read_until(message_terminator=b"\n", timeout=self.ADJUSTIK_TIMEOUT),
            call.read_until(message_terminator=b"\n", timeout=self.ADJUSTIK_TIMEOUT)
        ])



#! /usr/bin/env python3

import qmi, time
from qmi.instruments.raspberry.relay import RaspberryPiRelay

qmi.start("relay_server", "qmi.conf")

relay = qmi.make_instrument("relay", RaspberryPiRelay, pin_nr=7)

time.sleep(0.100)
input("\nRelay server active, press Enter to quit ...\n")



"""
Instrument driver for the Thorlabs PM100D and compatible optical power meter.
"""

import logging
import re
import time
from typing import List, NamedTuple, Tuple

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.transport import create_transport, list_usbtmc_transports, UsbTmcTransportDescriptorParser

# Global variable holding the logger for this module.

_logger = logging.getLogger(__name__)

# Tuple type returned by get_sensor_info().
SensorInfo = NamedTuple('SensorInfo',
                        [('name', str),
                         ('serial', str),
                         ('cal_msg', str),
                         ('type', str),
                         ('subtype', str),
                         ('flags', int)])


class Thorlabs_PM10x(QMI_Instrument):
    """General instrument driver for the Thorlabs PM10x optical power meters.

    This driver is also compatible with the Thorlabs PM16-120.
    """
    VENDOR_ID = 0x1313  # Thorlabs' vendor ID.
    PRODUCT_ID = 0x0000  # To be overridden by deriving subclasses.

    # Default response timeout in seconds.
    DEFAULT_RESPONSE_TIMEOUT = 5.0

    @staticmethod
    def list_instruments() -> List[str]:
        """Return a list of QMI transport descriptors for attached power meters."""
        devices = []
        parser = UsbTmcTransportDescriptorParser
        for desc_str in list_usbtmc_transports():
            if parser.match_interface(desc_str):
                parameters = parser.parse_parameter_strings(desc_str)
                if parameters.get("vendorid") == Thorlabs_PM100D.VENDOR_ID and \
                        parameters.get("productid") in (Thorlabs_PM100D.PRODUCT_ID,
                                                        Thorlabs_PM100USB.PRODUCT_ID,
                                                        Thorlabs_PM101U.PRODUCT_ID,
                                                        Thorlabs_PM16_120.PRODUCT_ID):
                    devices.append(desc_str)
        return devices

    def __init__(self,
                 context: QMI_Context,
                 name: str,
                 transport: str
                 ) -> None:
        super().__init__(context, name)
        self._timeout = self.DEFAULT_RESPONSE_TIMEOUT
        self._transport = create_transport(transport)
        self._scpi_protocol = ScpiProtocol(self._transport, default_timeout=self._timeout)

    @rpc_method
    def open(self) -> None:
        _logger.info("[%s] Opening connection to instrument", self._name)
        self._check_is_closed()
        self._transport.open()
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("[%s] Closing connection to instrument", self._name)
        super().close()
        self._transport.close()

    def _check_error(self) -> None:
        """Read the instrument error queue and raise an exception if there is an error."""
        resp = self._scpi_protocol.ask("SYST:ERR?")
        # When there are no errors, the response is '+0,"No error"'.
        if not re.match(r"^\s*[-+]?\s*0\s*,", resp):
            # Some error occurred.
            raise QMI_InstrumentException("Instrument returned error: {}".format(resp))

    @rpc_method
    def reset(self) -> None:
        """Reset the instrument, returning (most) settings to their defaults."""
        self._check_is_open()
        self._scpi_protocol.write("*CLS")  # clear error queue
        self._scpi_protocol.write("*RST")
        self._scpi_protocol.ask("*OPC?")
        self._check_error()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance."""
        self._check_is_open()
        resp = self._scpi_protocol.ask("*IDN?")
        words = resp.rstrip().split(",")
        if len(words) != 4:
            raise QMI_InstrumentException("Unexpected response to *IDN?, got {!r}".format(resp))
        return QMI_InstrumentIdentification(vendor=words[0],
                                            model=words[1],
                                            serial=words[2],
                                            version=words[3])

    @rpc_method
    def get_sensor_info(self) -> SensorInfo:
        """Get sensor type and return information as a SensorInfo instance."""
        self._check_is_open()
        resp = self._scpi_protocol.ask('SYST:SENS:IDN?')
        words = resp.split(',')
        if len(words) != 6:
            raise QMI_InstrumentException("Unexpected response to SYST:SENS:IDN?, got {!r}".format(resp))
        return SensorInfo(name=words[0],
                          serial=words[1],
                          cal_msg=words[2],
                          type=words[3],
                          subtype=words[4],
                          flags=int(words[5]))

    @rpc_method
    def get_timestamped_power(self) -> Tuple[float, float]:
        """Start measurement and return measured power [W]."""
        self._check_is_open()
        resp = self._scpi_protocol.ask('MEAS:POW?')
        timestamp = time.time()
        power = float(resp)
        return timestamp, power

    @rpc_method
    def get_power(self) -> float:
        """Start measurement and return measured power [W].

        This function is part of a generic power meter API
        and must be supported by all power meter implementations.
        """
        (_, power) = self.get_timestamped_power()
        return power

    @rpc_method
    def get_range(self) -> float:
        """Get measurement range [W]."""
        self._check_is_open()
        resp = self._scpi_protocol.ask("SENS:POW:RANG?")
        return float(resp)

    @rpc_method
    def get_range_in_use(self) -> float:
        """Return the currently used measuring range in Watt.

        This function is part of a generic power meter API
        and must be supported by all power meter implementations.
        Other implementations may return the range as an index or code.
        """
        return self.get_range()

    @rpc_method
    def set_range(self, measurement_range: float) -> None:
        """Set measurement range [W]."""
        self._check_is_open()
        cmd = "SENS:POW:RANG {}".format(measurement_range)
        self._scpi_protocol.write(cmd)
        self._check_error()

    @rpc_method
    def get_autorange(self) -> bool:
        """Return True if the power meter is in auto range mode."""
        self._check_is_open()
        resp = self._scpi_protocol.ask("SENS:POW:RANG:AUTO?")
        return bool(int(resp))

    @rpc_method
    def set_autorange(self, enable: bool) -> None:
        """Enable or disable auto range mode.

        This function is part of a generic power meter API
        and must be supported by all power meter implementations.
        """
        self._check_is_open()
        value = 1 if enable else 0
        cmd = "SENS:POW:RANG:AUTO {}".format(value)
        self._scpi_protocol.write(cmd)
        self._check_error()

    @rpc_method
    def get_wavelength(self) -> float:
        """Return configured wavelength [nm]."""
        self._check_is_open()
        resp = self._scpi_protocol.ask('SENS:CORR:WAV?')
        return float(resp)

    @rpc_method
    def set_wavelength(self, wavelength: float) -> None:
        """Configure the wavelength in nm.

        This function is part of a generic power meter API
        and must be supported by all power meter implementations.
        """
        self._check_is_open()
        cmd = "SENS:CORR:WAV {}".format(wavelength)
        self._scpi_protocol.write(cmd)
        self._check_error()

    @rpc_method
    def get_current(self) -> float:
        """Start measurement and return measured diode current in Ampere."""
        resp = self._scpi_protocol.ask('MEAS:CURR?')
        return float(resp)


class Thorlabs_PM100D(Thorlabs_PM10x):
    """Alias class for PM100D; uses the interface of PM10x."""
    PRODUCT_ID = 0x8078


class Thorlabs_PM100USB(Thorlabs_PM10x):
    """Alias class for PM100USB; uses the interface of PM10x."""
    PRODUCT_ID = 0x8072


class Thorlabs_PM101U(Thorlabs_PM10x):
    """Alias class for PM101U; uses the interface of PM10x."""
    PRODUCT_ID = 0x8076


class Thorlabs_PM16_120(Thorlabs_PM10x):
    """Alias class for PM16_120; uses the interface of PM10x."""
    PRODUCT_ID = 0x807b



"""
Instrument driver for IP Power relay unit.
"""

from enum import Enum
from functools import partial
from ipaddress import IPv4Address
import logging
import re
import time
from typing import Dict, Optional
import urllib.error
import urllib.parse
import urllib.request

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException, QMI_TimeoutException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method


# Logger for this module
_logger = logging.getLogger(__name__)


# Default timeout for HTTP requests (seconds)
TIMEOUT = 5


class PowerSocket(Enum):
    """Socket identifiers."""
    P1 = "p61"
    P2 = "p62"
    P3 = "p63"
    P4 = "p64"


class Command(Enum):
    """Commands."""
    GETVERSION = "getversion"
    GETMAC = "getmac"
    GETPOWER = "getpower"
    SETPOWER = "setpower"
    CYCLEPOWER = "setpowercycle"


class PowerState(Enum):
    """Power state (on/off) for sockets."""
    OFF = 0
    ON = 1


def _parse_response(response):
    # Response is always in the form:
    #     <!--CGI-DATABEG-->\n<p>\n...</p>\n\n<!--CGI-DATAEND-->
    # We could use html.parser, but that seems overkill.
    payload = response.read()
    match = re.search(r"<p>(.*)</p>", payload.decode('utf-8'), re.DOTALL)

    if match is None:
        raise ValueError("Invalid response.")

    try:
        message = match.group(1).strip()
    except IndexError as exc:
        raise ValueError("Invalid response.") from exc

    return message


def _parse_status_string(status_string):
    # Status string is of the form p6x=X, ...
    try:
        state_strings = re.findall(r"(p6[0-9]{1}=[01]{1})", status_string)
        states = {PowerSocket(p): PowerState(int(s)) for p, s in map(partial(str.split, sep='='), state_strings)}
        # Also no match (empty dictionary 'states') should raise the exception
        if len(states) == 0:
            raise Exception

    except Exception as exc:
        raise QMI_InstrumentException("Instrument status string is invalid.") from exc

    return states


class IPPower9850(QMI_Instrument):
    """QMI driver for the IP Power relay unit."""

    def __init__(self,
                 context: QMI_Context,
                 name: str,
                 ipv4_address: str,
                 username: str = "admin",
                 password: str = "12345678") -> None:
        """Initialize the instrument driver.

        Parameters:
            context:        QMI Context.
            name:           Identifier for the instrument in the QMI context.
            ipv4_address:   IPv4 address of the device on the network.
            username:       Username (factory default: 'admin')
            password:       Associated password (factory default: '12345678')
        """
        super().__init__(context, name)

        self._host = IPv4Address(ipv4_address)  # raises ValueError for improper address
        self._baseurl = f"http://{self._host}"
        self._username = username
        self._password = password
        self._opener = None  # type: Optional[urllib.request.OpenerDirector]

    def _send_command(self, command, **kwargs):
        if self._opener is None:
            raise RuntimeError("Open device first")

        # Build the query url.
        params = {"cmd": command.value}
        for field, value in kwargs.items():
            params[field] = value
        url = "{}/set.cmd?{}".format(self._baseurl, urllib.parse.urlencode(params))
        _logger.debug("Request: %s", url)

        # Post request and check for response.
        req = urllib.request.Request(url, method="GET")
        try:
            with self._opener.open(req, timeout=TIMEOUT) as response:
                answer = _parse_response(response)
        except urllib.error.HTTPError as exc:
            raise QMI_InstrumentException("Error in communication with device") from exc

        return answer

    @rpc_method
    def open(self) -> None:
        """Open the device interface by issuing a HEAD request and check for a response."""
        _logger.info("Opening connection to IPPower9850 at %s", self._host)

        # Install a basic authentication handlers.
        auth_handler = urllib.request.HTTPBasicAuthHandler()
        auth_handler.add_password(realm="guest area",  # type: ignore  # defined in AbstractBasicAuthHandler.__init__()
                                  uri=self._baseurl,
                                  user=self._username,
                                  passwd=self._password)
        auth_handler.add_password(realm="user area",  # type: ignore  # defined in AbstractBasicAuthHandler.__init__()
                                  uri=self._baseurl,
                                  user=self._username,
                                  passwd=self._password)
        self._opener = urllib.request.build_opener(auth_handler)

        # Perform request and check for 200 OK response
        req = urllib.request.Request(self._baseurl, method="HEAD")
        try:
            resp = self._opener.open(req, timeout=TIMEOUT)
        except urllib.error.HTTPError as exc:
            raise QMI_InstrumentException("Access denied for device at {}".format(self._host)) from exc
        except urllib.error.URLError as exc:
            raise QMI_TimeoutException("Unable to find a device at {}".format(self._host)) from exc

        if resp.status != 200:
            # This should not happen - exception should have been raised by urllib
            raise QMI_InstrumentException("Unknown error in accessing device at {}".format(self._host))

        return super().open()

    @rpc_method
    def close(self) -> None:
        """Close the device interface."""
        # Nothing to be done.
        return super().close()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Retrieve the device identification.

        The MAC address of the device is used as unique serial number. The version field holds the firmware version.
        """
        # Fixed vendor and model, no way to query it
        vendor = "Aviosys"
        model = "IP Power 9850XX"  # XX denotes socket type; for Schuko it is GE

        # Use MAC address as serial number (response is of the form 'mac=...' with no formatting (just 12 letters))
        mac = self._send_command(Command.GETMAC).split('=')[1]
        mac_formatted = ':'.join(mac[i:i+2] for i in range(0, 12, 2))

        # Firmware version (response is of the form 'Version=...')
        version = self._send_command(Command.GETVERSION).split('=')[1]  # firmware version
        return QMI_InstrumentIdentification(vendor=vendor, model=model, serial=mac_formatted, version=version)

    @rpc_method
    def get_all_states(self) -> Dict[PowerSocket, PowerState]:
        """Retrieve the power state of all channels."""
        response_message = self._send_command(Command.GETPOWER)
        return _parse_status_string(response_message)

    @rpc_method
    def get_state(self, channel: PowerSocket) -> PowerState:
        """Retrieve the power state of a single specific channel.

        Parameters:
            channel:    power socket identifier.
        """
        all_states = self.get_all_states()
        return all_states[channel]

    @rpc_method
    def set_state(self, channel: PowerSocket, target_state: PowerState) -> bool:
        """Set the power state for a single specified channel.

        Parameters:
            channel:        power socket identifier.
            target_state:   target state (on/off).

        Returns:
            True if state was set successfully, False otherwise.
        """
        args = {channel.value: target_state.value}
        status_string = self._send_command(Command.SETPOWER, **args)
        new_state = _parse_status_string(status_string)
        return new_state[channel] == target_state

    @rpc_method
    def set_states(self, target_states: Dict[PowerSocket, PowerState]) -> bool:
        """Set the power state of multiple channels at once.

        Parameters:
            target_states:  mapping of channels to target states (on/off).

        Returns:
            True if all states were set successfully, False otherwise.
        """
        args = {channel.value: target_state.value for channel, target_state in target_states.items()}
        status_string = self._send_command(Command.SETPOWER, **args)
        new_states = _parse_status_string(status_string)
        return all(new_states[p] == target_states[p] for p in target_states.keys())

    @rpc_method
    def set_all_off(self) -> bool:
        """Turn off all channels.

        Returns:
            True if all states were set off successfully, False otherwise.
        """
        target_states = {channel: PowerState.OFF for channel in PowerSocket}
        return self.set_states(target_states)

    @rpc_method
    def set_all_on(self) -> bool:
        """Turn on all channels.

        Returns:
            True if all states were set on successfully, False otherwise.
        """
        target_states = {channel: PowerState.ON for channel in PowerSocket}
        return self.set_states(target_states)

    @rpc_method
    def cycle(self, channel: PowerSocket, wait: int = 1, *, block: bool = False) -> bool:
        """Cycle power.

        For the specified channel the power is turned off and then, after `wait` seconds, turned back on.

        Note: the device returns OK when the command is accepted, so an additional check is needed to see if the
        cycle is actually successful. If `block` is True the routine will block until it verifies that the power is
        back on. It will block for at most `wait` + `ippower.TIMEOUT` seconds.

        Parameters:
            channel:    power socket identifier.
            wait:       wait time in seconds.
            block:      block until power cycling is verified.

        Returns:
            True if all channels cycled ok, False otherwise.
        """
        args = {channel.value: int(wait)}
        status_string = self._send_command(Command.CYCLEPOWER, **args)

        # Return message string is of the form p6x cycle ok, ...
        try:
            state_strings = re.findall(r"(p6[0-9]{1} cycle ok)", status_string)
            cycle_accepted = [PowerSocket(s[0:3]) for s in state_strings]
        except Exception as exc:
            raise QMI_InstrumentException("Instrument status string is invalid.") from exc

        accepted = channel in cycle_accepted
        cycle_ok = False
        if accepted and block:
            time.sleep(wait)
            t_start = time.monotonic()
            while time.monotonic() - t_start < TIMEOUT:
                time.sleep(0.1)
                if self.get_state(channel) == PowerState.ON:
                    cycle_ok = True
                    break
        else:
            cycle_ok = accepted

        return cycle_ok



#! /usr/bin/env python3

"""Test start/stop functionality of QMI framework with config files and optional inputs.
"""
import os
import unittest

ORIGINAL_QMI_CONFIG = os.getenv("QMI_CONFIG")
qmi_config_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "qmi.conf")
os.environ["QMI_CONFIG"] = qmi_config_path

import qmi
import qmi.core.context_singleton
import qmi.core.exceptions

QMI_ENV_CONFIG_FILE_PRESENT = False
QMI_ENV_CONFIG_FILE = os.path.join(str(os.getenv("QMI_CONFIG")), "qmi.conf")
if os.path.isfile(QMI_ENV_CONFIG_FILE):
    QMI_ENV_CONFIG_FILE_PRESENT = True

CONFIG_FILE_CONTEXT = """{
    # Log level for messages to the console.
    "logging": {
        "console_loglevel": "INFO"
    },

    "contexts": {
        # Testing remote instrument access.
        "instr_server": {
            "host": "127.0.0.1",
            "tcp_server_port": 40001
        },
        "instr_client": {
            "connect_to_peers": ["instr_server"]
        }
    }
}"""

CONTEXT_CFG = {"instr_server": {
        "host": 'localhost',
        "tcp_server_port": 10004,
        "connect_to_peers": [],
        "enabled": True,
        "program_module": "some_module",
        "program_args": [],
        "python_path": "SOMEPATH",
        "virtualenv_path": None
    },
    "instr_client": {
        "host": None,
        "tcp_server_port": None,
        "connect_to_peers": ["instr_server"],
        "enabled": False,
        "program_module": None,
        "program_args": ["arg1", "arg2"],
        "python_path": None,
        "virtualenv_path": "MOSEPATH"
        }
    }

BAD_OPTIONAL_CONTEXT_1 = {"instr_server": {
        "tcp_server_port": "10004"
    }}
BAD_OPTIONAL_CONTEXT_2 = {"instr_client": {
        "connect_to_peers": "instr_server"
        }
    }
BAD_OPTIONAL_CONTEXT_3 = {"instr_server": {
        "enabled": "True"
    }}


class TestContextConfigFileInputs(unittest.TestCase):

    def setUp(self) -> None:
        self.ctx_name = "hello_world"

    def tearDown(self):
        # Stop context if present.
        if self.ctx_name in qmi.info():
            qmi.stop()

        # Set these back to None to avoid false leads for later unit-tests
        qmi.core.context_singleton.QMI_CONFIG = None

    def test_01_no_input_no_environment(self):
        # Check that not giving an input file leads to use of default context.
        qmi.core.context_singleton.QMI_CONFIG = None

        qmi.start(self.ctx_name, None)
        ctx = qmi.context()

        self.assertEqual(self.ctx_name, ctx.name)
        self.assertTrue(ctx.get_config().config_file is None)
        self.assertTrue(ctx.get_config().qmi_home is None)

    def test_02_no_input_QMI_CONFIG_environment_set(self):
        # Check that, when QMI_CONFIG environment variable is "set", it is used if no input file given.
        # Add extra check that we do not accidentally overwrite/delete a real QMI_CONFIG qmi.conf file
        qmi_config_path = os.path.dirname(os.path.abspath(__file__))
        qmi_config_file = os.path.join(qmi_config_path, "qmi.conf")
        delete_later = False
        if not os.path.isfile(qmi_config_file):
            # First create a temporary qmi.conf file.
            delete_later = True
            with open(qmi_config_file, 'w') as qmi_conf:
                qmi_conf.write(CONFIG_FILE_CONTEXT)

        qmi.core.context_singleton.QMI_CONFIG = qmi_config_file

        qmi.start(self.ctx_name)
        ctx = qmi.context()
        self.assertEqual(ctx.name, self.ctx_name)
        self.assertEqual(ctx.get_config().config_file, qmi_config_file)
        self.assertTrue(ctx.get_config().qmi_home is None)

        if delete_later:
            os.remove(qmi_config_file)

    def test_03_input_config_file(self):
        # Check that context creation works with an input for input file.
        # First create a temporary qmi.conf file.
        qmi_config_path = os.path.dirname(os.path.abspath(__file__))
        qmi_config_file = os.path.join(qmi_config_path, "qmi.conf")
        with open(qmi_config_file, 'w') as qmi_conf:
            qmi_conf.write(CONFIG_FILE_CONTEXT)

        qmi.start(self.ctx_name, qmi_config_file)
        ctx = qmi.context()
        self.assertEqual(ctx.name, self.ctx_name)
        self.assertEqual(ctx.get_config().config_file, qmi_config_file)
        self.assertTrue(ctx.get_config().qmi_home is None)

        os.remove(qmi_config_file)

    def test_04_input_config_file_error(self):
        # Check that an error is raised when there is a mistake in the input file name/path
        qmi_config_path = os.path.dirname(os.path.abspath(__file__))
        qmi_config_file = os.path.join(qmi_config_path, "qmi.conf")

        with self.assertRaises(FileNotFoundError):
            qmi.start(self.ctx_name, qmi_config_file)

    def test_05_no_input_override_default_loglevel(self):
        # Check that we can override default loglevel.
        loglevel = "WARN"
        qmi.core.context_singleton.QMI_CONFIG = None

        qmi.start(self.ctx_name, console_loglevel=loglevel)
        ctx = qmi.context()

        self.assertEqual(ctx.get_config().logging.console_loglevel, loglevel.upper())

    def test_06_no_input_invalid_override_loglevel(self):
        # Check that we raise an error with an invalid loglevel.
        loglevel = "SILENT"
        qmi.core.context_singleton.QMI_CONFIG = None
        with self.assertRaises(qmi.core.exceptions.QMI_ConfigurationException):
            qmi.start(self.ctx_name, console_loglevel=loglevel)


class TestContextOptionalConfigInputs(unittest.TestCase):

    def test_01_no_conf_file_only_optional_cfg_input(self):
        # Check that not giving an input config file, but giving optional cfg input leads to use of optional context.
        context_name = "instr_server"
        second_context = "instr_client"
        qmi.core.context_singleton.QMI_CONFIG = None

        qmi.start(context_name, context_cfg=CONTEXT_CFG)

        try:
            contexts = qmi.get_configured_contexts()
            # Check first the server context values are correct
            context_server = contexts[context_name]
            input_server_context = CONTEXT_CFG[context_name]
            self.assertTrue(context_server.host is input_server_context["host"])
            self.assertTrue(context_server.tcp_server_port is input_server_context["tcp_server_port"])
            self.assertTrue(context_server.connect_to_peers == input_server_context["connect_to_peers"])
            self.assertTrue(context_server.enabled is input_server_context["enabled"])
            self.assertTrue(context_server.program_module is input_server_context["program_module"])
            self.assertTrue(context_server.program_args == input_server_context["program_args"])
            self.assertTrue(context_server.python_path is input_server_context["python_path"])
            self.assertTrue(context_server.virtualenv_path is input_server_context["virtualenv_path"])
            # Then check the client context values are correct
            context_client = contexts[second_context]
            input_client_context = CONTEXT_CFG[second_context]
            self.assertTrue(context_client.host is input_client_context["host"])
            self.assertTrue(context_client.tcp_server_port is input_client_context["tcp_server_port"])
            self.assertTrue(context_client.connect_to_peers == input_client_context["connect_to_peers"])
            self.assertTrue(context_client.enabled is input_client_context["enabled"])
            self.assertTrue(context_client.program_module is input_client_context["program_module"])
            self.assertTrue(context_client.program_args == input_client_context["program_args"])
            self.assertTrue(context_client.python_path is input_client_context["python_path"])
            self.assertTrue(context_client.virtualenv_path is input_client_context["virtualenv_path"])

        finally:
            qmi.stop()

    def test_02_conf_file_context_overriden_by_optional_cfg_input(self):
        # Check that the initial config values read from config file get overridden with optional config input
        # First create a temporary qmi.conf file.
        qmi_config_path = os.path.dirname(os.path.abspath(__file__))
        qmi_config_file = os.path.join(qmi_config_path, "qmi.conf")
        with open(qmi_config_file, 'w') as qmi_conf:
            qmi_conf.write(CONFIG_FILE_CONTEXT)

        context_name = "instr_server"
        second_context = "instr_client"

        qmi.start(context_name, qmi_config_file, context_cfg=CONTEXT_CFG)
        try:
            contexts = qmi.get_configured_contexts()
            # Check first the server context values are correct
            context_server = contexts[context_name]
            input_server_context = CONTEXT_CFG[context_name]
            self.assertTrue(context_server.host is input_server_context["host"])
            self.assertTrue(context_server.tcp_server_port is input_server_context["tcp_server_port"])
            self.assertTrue(context_server.connect_to_peers == input_server_context["connect_to_peers"])
            self.assertTrue(context_server.enabled is input_server_context["enabled"])
            self.assertTrue(context_server.program_module is input_server_context["program_module"])
            self.assertTrue(context_server.program_args == input_server_context["program_args"])
            self.assertTrue(context_server.python_path is input_server_context["python_path"])
            self.assertTrue(context_server.virtualenv_path is input_server_context["virtualenv_path"])
            # Then check the client context values are correct
            context_client = contexts[second_context]
            input_client_context = CONTEXT_CFG[second_context]
            self.assertTrue(context_client.host is input_client_context["host"])
            self.assertTrue(context_client.tcp_server_port is input_client_context["tcp_server_port"])
            self.assertTrue(context_client.connect_to_peers == input_client_context["connect_to_peers"])
            self.assertTrue(context_client.enabled is input_client_context["enabled"])
            self.assertTrue(context_client.program_module is input_client_context["program_module"])
            self.assertTrue(context_client.program_args == input_client_context["program_args"])
            self.assertTrue(context_client.python_path is input_client_context["python_path"])
            self.assertTrue(context_client.virtualenv_path is input_client_context["virtualenv_path"])

        finally:
            qmi.stop()
            os.remove(qmi_config_file)

    def test_03_see_that_typecheck_catches_wrong_input_types(self):
        # Check that the type checking in context_singleton catches wrong types on inputs in optional config
        context_name = "instr_server"
        second_context = "instr_client"
        qmi.core.context_singleton.QMI_CONFIG = None

        with self.assertRaises(qmi.core.exceptions.QMI_ConfigurationException):
            qmi.start(context_name, context_cfg=BAD_OPTIONAL_CONTEXT_1)

        with self.assertRaises(qmi.core.exceptions.QMI_ConfigurationException):
            qmi.start(second_context, context_cfg=BAD_OPTIONAL_CONTEXT_2)

        with self.assertRaises(qmi.core.exceptions.QMI_ConfigurationException):
            qmi.start(context_name, context_cfg=BAD_OPTIONAL_CONTEXT_3)

    def test_04_no_context_started_raises_QMI_NoActiveContextException(self):
        with self.assertRaises(qmi.core.exceptions.QMI_NoActiveContextException):
            qmi.make_rpc_object("fail", object)

        with self.assertRaises(qmi.core.exceptions.QMI_NoActiveContextException):
            qmi.make_instrument("fail", object)

        with self.assertRaises(qmi.core.exceptions.QMI_NoActiveContextException):
            qmi.make_task("fail", object)

        with self.assertRaises(qmi.core.exceptions.QMI_NoActiveContextException):
            qmi.get_rpc_object("fail")

        with self.assertRaises(qmi.core.exceptions.QMI_NoActiveContextException):
            qmi.get_instrument("fail")

        with self.assertRaises(qmi.core.exceptions.QMI_NoActiveContextException):
            qmi.get_task("fail")


if __name__ == "__main__":
    unittest.main()
    os.environ["QMI_CONFIG"] = ORIGINAL_QMI_CONFIG



"""
Pololu Robotics & Electronics, motion control modules.

The qmi.instruments.pololu package provides support for:
- Maestro USB servo controllers.
"""

from qmi.instruments.pololu.maestro import Pololu_Maestro



"""Read and process configuration files.

The QMI configuration file language closely resembles JSON.
The configuration language differs from JSON as follows:

- The top-level value in the file must be a dictionary
  (the JSON specification calls this an `object`).
- Comments are allowed (and ignored during reading).
- Comments start with ``#`` and continue until the end of the line.
- Comments can not be placed inside strings.
"""

import collections
import json
import re

from qmi.core.exceptions import QMI_ConfigurationException


def _strip_comments(s: str) -> str:
    """Strip comments from a JSON-like string."""

    # Match first # character (excluding strings).
    re_comment = re.compile(r'^(?:[^"#]|"(?:[^\\"]|\\.)*")*#')

    # Split into lines.
    lines = re.split(r'[\r\n]', s)

    # Process each line.
    for (i, line) in enumerate(lines):
        # Strip comments from this line.
        if "#" in line:
            m = re_comment.match(line)
            if m is not None:
                # Found comment. Strip it.
                lines[i] = line[:m.end()-1]

    # Join lines to form stripped string.
    return "\n".join(lines)


def config_pairs_hook(config_pairs) -> collections.OrderedDict:
    """Reject duplicate keys."""
    cfg_dict = collections.OrderedDict()
    for k, v in config_pairs:
        if k in cfg_dict:
            raise ValueError("duplicate key: {}".format(k))

        else:
            cfg_dict[k] = v

    return cfg_dict


def load_config_string(s: str) -> dict:
    """Load configuration data from a string.

    Parameters:
        s: String containing configuration data.

    Returns:
        Parsed configuration data checked for duplicates.
    """

    s = _strip_comments(s)
    cfg = json.loads(s, object_pairs_hook=config_pairs_hook)
    if not isinstance(cfg, dict):
        raise QMI_ConfigurationException("Expecting mapping at top level of configuration but got {}"
                                         .format(type(cfg).__name__))

    return cfg


def load_config_file(filename: str) -> dict:
    """Load configuration data from a file.

    Parameters:
        filename: Path of configuration file.

    Returns:
        Parsed configuration data.
    """
    with open(filename, "r") as f:
        s = f.read()
    return load_config_string(s)


def dump_config_string(cfg: dict) -> str:
    """Serialize configuration data to a string.

    Parameters:
        cfg: Configuration data.

    Returns:
        Serialized configuration string.
    """

    if not isinstance(cfg, dict):
        raise QMI_ConfigurationException("Expecting mapping at top level of configuration but got {}"
                                         .format(type(cfg).__name__))

    return json.dumps(cfg, indent=4)


def dump_config_file(cfg: dict, filename: str) -> None:
    """Write configuration data to a file.

    Parameters:
        cfg: Configuration data.
        filename: Path of configuration file to write.
    """
    s = dump_config_string(cfg)
    with open(filename, "w") as f:
        f.write(s)



"""Testcases of Quantum opus AmpSim Module instrument"""
import unittest
from unittest.mock import call, MagicMock

import qmi
from qmi.instruments.stanford_research_systems import Srs_Sim900
from qmi.instruments.quantum_opus import QuantumOpus_QoAmpSim
from qmi.core.exceptions import QMI_UsageException, QMI_InstrumentException


class TestAmpSimModule(unittest.TestCase):
    """Testcase of Quantum opus AmpSim Module instrument"""

    def setUp(self):
        qmi.start("test_quantum_opus_ampsim900")
        self._sim900 = MagicMock(spec=Srs_Sim900)
        self._port = 6
        self.amp_sim_module = qmi.make_instrument("QOAmpSim", QuantumOpus_QoAmpSim, self._sim900, self._port)

    def tearDown(self):
        qmi.stop()

    def test_get_module_id(self):
        # arrange
        port = 6
        expected_value = 'howdyhi'
        self._sim900.ask_module.return_value = expected_value
        expected_calls = [
            call.ask_module(port, "+A?")
            ]
        # act
        bias_current = self.amp_sim_module .get_module_id()
        # assert
        self.assertEqual(self._sim900.method_calls, expected_calls)
        self.assertEqual(bias_current, expected_value)

    def test_get_device_bias_current(self):
        # arrange
        self._sim900.ask_module.return_value = '60000'
        expected_calls = [
            call.ask_module(self._port, "+B?")
            ]
        # act
        self.amp_sim_module.get_device_bias_current()
        # assert
        self.assertEqual(self._sim900.method_calls, expected_calls)

    def test_set_device_bias_current(self):
        # arrange
        bias_current = 19000
        # act
        self.amp_sim_module.set_device_bias_current(bias_current)
        # assert
        self._sim900.send_terminated_message.assert_called_once_with(6, '+B19000;')
        
    def test_set_device_bias_current_raise_out_of_range(self):
        with self.assertRaises(QMI_UsageException):
            self.amp_sim_module.set_device_bias_current(-1)
        with self.assertRaises(QMI_UsageException):
            self.amp_sim_module.set_device_bias_current(70000)

    def test_set_adc_low_gain_mode(self):
        # act
        self.amp_sim_module.set_adc_low_gain_mode()
        # assert
        self._sim900.send_terminated_message.assert_called_once_with(self._port, '+C1;')

    def test_set_adc_high_gain_mode(self):
        # act
        self.amp_sim_module.set_adc_high_gain_mode()
        # assert
        self._sim900.send_terminated_message.assert_called_once_with(self._port, '+C0;')

    def test_get_device_voltage(self):
        # arrange
        self._sim900.ask_module.return_value = '60000'
        expected_calls = [
            call.ask_module(self._port, "+C?")
            ]
        # act
        result = self.amp_sim_module.get_device_voltage()
        # assert
        self.assertEqual(self._sim900.method_calls, expected_calls)
        self.assertEqual(result, 60000)

    def test_set_reset_event_duration(self):
        # arrange
        event_duration = 10
        # act
        self.amp_sim_module.set_reset_event_duration(event_duration)
        # assert
        self._sim900.send_terminated_message.assert_called_once_with(6, '+D10;')

    def test_set_reset_event_duration_out_of_range(self):
        with self.assertRaises(QMI_UsageException):
            self.amp_sim_module.set_reset_event_duration(-1)
        with self.assertRaises(QMI_UsageException):
            self.amp_sim_module.set_reset_event_duration(256)

    def test_get_reset_event_duration(self):
        # arrange
        self._sim900.ask_module.return_value = '10'
        expected_calls = [
            call.ask_module(self._port, "+D?")
            ]
        # act
        result = self.amp_sim_module.get_reset_event_duration()
        # assert
        self.assertEqual(self._sim900.method_calls, expected_calls)
        self.assertEqual(result, 10)

    def test_set_auto_reset_enabled(self):
        # act
        self.amp_sim_module.set_auto_reset_enabled(True)
        # assert
        self._sim900.send_terminated_message.assert_called_once_with(self._port, '+E1;')

    def test_set_auto_reset_disabled(self):
        # act
        self.amp_sim_module.set_auto_reset_enabled(False)
        # assert
        self._sim900.send_terminated_message.assert_called_once_with(self._port, '+E0;')

    def test_get_auto_reset_enabled(self):
        # arrange
        self._sim900.ask_module.return_value = '1'
        expected_calls = [
            call.ask_module(self._port, "+E?")
            ]
        # act
        result = self.amp_sim_module.get_auto_reset_enabled()
        # assert
        self.assertEqual(self._sim900.method_calls, expected_calls)
        self.assertEqual(result, True)

    def test_get_auto_reset_disabled(self):
        # arrange
        self._sim900.ask_module.return_value = '0'
        expected_calls = [
            call.ask_module(self._port, "+E?")
            ]
        # act
        result = self.amp_sim_module.get_auto_reset_enabled()
        # assert
        self.assertEqual(self._sim900.method_calls, expected_calls)
        self.assertEqual(result, False)

    def test_initiate_reset_event(self):
        # act
        self.amp_sim_module.initiate_reset_event()
        # assert
        self._sim900.send_terminated_message.assert_called_once_with(self._port, '+F;')

    def test_initiate_auto_bias_function(self):
        # act
        self.amp_sim_module.initiate_auto_bias_function()
        # assert
        self._sim900.send_terminated_message.assert_called_once_with(self._port, '+G;')

    def test_store_bias_current_in_non_volatile_memory(self):
        # act
        self.amp_sim_module.store_bias_current_in_non_volatile_memory(60000)
        # assert
        self._sim900.send_terminated_message.assert_called_once_with(self._port, '+H60000;')

    def test_store_bias_current_in_non_volatile_memory_raise_out_of_range(self):
        with self.assertRaises(QMI_UsageException):
            self.amp_sim_module.store_bias_current_in_non_volatile_memory(-1)
        with self.assertRaises(QMI_UsageException):
            self.amp_sim_module.store_bias_current_in_non_volatile_memory(70000)

    def test_get_bias_current_from_non_volatile_memory(self):
        # arrange
        self._sim900.ask_module.return_value = '19000,0\n'
        expected_calls = [
            call.ask_module(self._port, "+H?;")
            ]
        # act
        bias_current = self.amp_sim_module.get_bias_current_from_non_volatile_memory()
        # assert
        self.assertEqual(self._sim900.method_calls, expected_calls)
        self.assertEqual(bias_current, 19000)

    def test_get_bias_current_from_non_volatile_memory_expect_raise(self):
        # arrange
        self._sim900.ask_module.return_value = '-19000'
        # act
        # assert
        with(self.assertRaises(QMI_InstrumentException)):
            self.amp_sim_module.get_bias_current_from_non_volatile_memory()

    def test_use_device_bias_from_non_volatile_memory(self):
        # act
        self.amp_sim_module.use_device_bias_from_non_volatile_memory()
        # assert
        self._sim900.send_terminated_message.assert_called_once_with(self._port, '+H;')



"""Instrument driver for the Cobolt CW Laser
"""

import logging

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class Cobolt_Laser_06_01(QMI_Instrument):
    """Instrument driver for the Cobolt 06-01 series diode laser."""

    _rpc_constants = ["FAULT_CODES", "OPERATING_MODES"]

    # Fault codes returned by get_operating_fault().
    FAULT_CODES = {
        0: "no errors",
        1: "temperature error",
        3: "interlock error",
        4: "constant power time out"
    }

    # Operating modes returned by get_operating_mode().
    OPERATING_MODES = {
        0: "Off",
        1: "Waiting for key",
        2: "Continuous",
        3: "On/Off Modulation",
        4: "Modulation",
        5: "Fault",
        6: "Aborted"
    }

    # Instrument should respond within 2 seconds.
    RESPONSE_TIMEOUT = 2.0

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        super().__init__(context, name)
        self._transport = create_transport(transport, default_attributes={"baudrate": 115200})

    @rpc_method
    def open(self) -> None:
        _logger.info("Opening connection to %s", self._name)
        self._transport.open()
        self._transport.discard_read()
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("Closing connection to %s", self._name)
        super().close()
        self._transport.close()

    @rpc_method
    def reset(self) -> None:
        """Reset the instrument, returning (most) settings to their defaults."""
        _logger.info("Resetting %s", self._name)
        self._write("@cob1")

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance."""
        serial = self._ask("gsn?")
        return QMI_InstrumentIdentification(vendor="Cobolt",
                                            model="06-01",
                                            serial=serial,
                                            version=None)

    def _write(self, cmd: str) -> None:
        """Send command to instrument and check instrument responds with "OK"."""

        self._check_is_open()

        # Avoid reading stale response from instrument.
        self._transport.discard_read()

        # Send command.
        self._transport.write(cmd.encode('ascii') + b"\r")

        # Read response.
        # NOTE: Manual says response is terminated by "\r",
        # but actual device terminates with "\r\n".
        resp = self._transport.read_until(message_terminator=b"\n",
                                          timeout=self.RESPONSE_TIMEOUT)
        if resp.rstrip() != b"OK":
            raise QMI_InstrumentException(
                "Expecting 'OK' response to command {!r} but got {!r}".format(cmd, resp))

    def _ask(self, cmd: str) -> str:
        """Send command to instrument and return response from instrument."""

        self._check_is_open()

        # Avoid reading stale response from instrument.
        self._transport.discard_read()

        # Send command.
        self._transport.write(cmd.encode('ascii') + b"\r")

        # Read response.
        resp = self._transport.read_until(message_terminator=b"\n",
                                          timeout=self.RESPONSE_TIMEOUT)

        return resp.rstrip().decode('ascii', errors='replace')

    @staticmethod
    def _parse_int(s: str) -> int:
        """Parse integer response from laser.

        Return integer value or raise QMI_InstrumentException.
        """
        try:
            return int(s)
        except ValueError:
            raise QMI_InstrumentException("Expecting integer from instrument but got {!r}".format(s))

    @staticmethod
    def _parse_float(s: str) -> float:
        """Parse floating point response from laser.

        Return float value or raise QMI_InstrumentException.
        """
        try:
            return float(s)
        except ValueError:
            raise QMI_InstrumentException("Expecting float from instrument but got {!r}".format(s))

    @staticmethod
    def _parse_bool(s: str) -> bool:
        """Parse boolean response from laser.

        Return response mapped to boolean value (0=False, 1=True) or raise QMI_InstrumentException.
        """
        try:
            v = int(s)
            if v == 0:
                return False
            elif v == 1:
                return True
        except ValueError:
            # error will be handled below
            pass
        raise QMI_InstrumentException("Expecting 0 or 1 from instrument but got {!r}".format(s))

    @rpc_method
    def get_laser_on_state(self) -> bool:
        """Return True if the laser is on, False if it is off."""
        resp = self._ask("l?")
        return self._parse_bool(resp)

    @rpc_method
    def set_laser_on_state(self, value: bool) -> None:
        """Set the laser on or off."""
        if value:
            _logger.info("Setting laser ON")
            cmd = "l1"
        else:
            _logger.info("Setting laser OFF")
            cmd = "l0"
        self._write(cmd)

    @rpc_method
    def get_operating_hours(self) -> float:
        """Get number of hours the laser head has operated."""
        resp = self._ask("hrs?")
        return self._parse_float(resp)

    @rpc_method
    def get_interlock_state(self) -> bool:
        """Return True if interlock is open, False if interlock allow laser on."""
        resp = self._ask("ilk?")
        return self._parse_bool(resp)

    @rpc_method
    def get_output_power_setpoint(self) -> float:
        """Get power set point [W]."""
        resp = self._ask("p?")
        return self._parse_float(resp)

    @rpc_method
    def set_output_power_setpoint(self, value: float) -> None:
        """ Set laser output power [W]."""
        self._write("p {:.6f}".format(value))

    @rpc_method
    def get_output_power(self) -> float:
        """Get actual laser output power [W]."""
        resp = self._ask("pa?")
        return self._parse_float(resp)

    @rpc_method
    def get_drive_current(self) -> float:
        """Get laser drive current [mA]"""
        resp = self._ask("i?")
        return self._parse_float(resp)

    @rpc_method
    def set_drive_current(self, value: float) -> None:
        """Set laser drive current [mA]."""
        self._write("slc {:.6f}".format(value))

    @rpc_method
    def set_constant_power_mode(self) -> None:
        """Enter constant power mode."""
        self._write("cp")

    @rpc_method
    def set_constant_current_mode(self) -> None:
        """Enter constant current mode."""
        self._write("ci")

    @rpc_method
    def get_fault(self) -> int:
        """Get operating fault state.

        See Cobolt_Laser_06_01.FAULT_CODES for the meaning of return codes.
        """
        resp = self._ask("f?")
        return self._parse_int(resp)

    @rpc_method
    def clear_fault(self) -> None:
        """Clear laser fault."""
        self._write("cf")

    @rpc_method
    def set_modulation_mode(self) -> None:
        """Enter modulatuion mode."""
        self._write("em")

    @rpc_method
    def get_digital_modulation_state(self) -> bool:
        """Return True if digital modulation is enabled."""
        resp = self._ask("gdmes?")
        return self._parse_bool(resp)

    @rpc_method
    def set_digital_modulation_state(self, value: bool) -> None:
        """Set digital modulation enable state."""
        v = 1 if value else 0
        self._write("sdmes {}".format(v))

    @rpc_method
    def get_analog_modulation_state(self) -> bool:
        """Get analog modulation enable state."""
        resp = self._ask("games?")
        return self._parse_bool(resp)

    @rpc_method
    def set_analog_modulation_state(self, value: bool) -> None:
        """Set analog modulation enable state."""
        v = 1 if value else 0
        self._write("sames {}".format(v))

    @rpc_method
    def get_operating_mode(self) -> int:
        """Get current operating mode.

        See Cobolt_Laser_06_01.OPERATING_MODES for the meaning of return codes.
        """
        resp = self._ask("gom?")
        return self._parse_int(resp)

    @rpc_method
    def get_analog_low_impedance_state(self) -> bool:
        """Return True if analog low impedance (50 Ohm) state is enabled."""
        resp = self._ask("galis?")
        return self._parse_bool(resp)

    @rpc_method
    def set_analog_low_impedance_state(self, value: bool) -> None:
        """Enable or disable analog low impedance (50 Ohm) state."""
        v = 1 if value else 0
        self._write("salis {}".format(v))

    @rpc_method
    def get_autostart_state(self) -> bool:
        """Return True if autostart is enabled.

        NOTE: This command is not in the manual. Private communication with Cobolt.
        """
        resp = self._ask("@cobas?")
        return self._parse_bool(resp)

    @rpc_method
    def set_autostart_state(self, value: bool) -> None:
        """Enable or disable autostart mode.

        NOTE: This command is not in the manual. Private communication with Cobolt.
        """
        v = 1 if value else 0
        self._write("@cobas {}".format(v))



"""Unit tests for Rigol DG4102 QMI driver."""

import unittest
from unittest.mock import MagicMock, call, patch

import logging
from typing import cast

import qmi
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.transport import QMI_TcpTransport
from qmi.instruments.rigol import Rigol_Dg4102
from qmi.utils.context_managers import open_close


class SuppressLogging:
    """Context manager to temporarily suppress logging during a test."""

    def __enter__(self):
        # Suppress logging of all levels up to ERROR.
        logging.getLogger("qmi.core.instrument").setLevel(logging.CRITICAL)

    def __exit__(self, typ, value, tb):
        # Restore default log levels.
        logging.getLogger("qmi.core.instrument").setLevel(logging.NOTSET)


class RigolDg4102TestCase(unittest.TestCase):

    def setUp(self):
        qmi.start("TestContext")
        self._transport_mock = MagicMock(spec=QMI_TcpTransport)
        with patch(
                'qmi.instruments.rigol.dg4102.create_transport',
                return_value=self._transport_mock):
            self.instr: Rigol_Dg4102 = qmi.make_instrument("instr", Rigol_Dg4102, "transport_descriptor")
            self.instr = cast(Rigol_Dg4102, self.instr)

    def tearDown(self):
        qmi.stop()

    def test_open_close(self):
        """Test that the open and close functions include calls as expected."""
        self.instr.open()
        self._transport_mock.open.assert_called_once_with()
        self._transport_mock.open.reset_mock()
        self._transport_mock.write.reset_mock()
        self.instr.close()
        self._transport_mock.close.assert_called_once_with()

    def test_set_source(self):
        """Test wait and clear function."""
        expected_init_source = 1
        new_source = 2
        with open_close(self.instr):
            source_init = self.instr.get_source()
            self.instr.set_source(2)
            source_new = self.instr.get_source()

        self.assertEqual(expected_init_source, source_init)
        self.assertEqual(new_source, source_new)

    def test_get_idn(self):
        """Test get_idn call"""
        vendor = "Rigol"
        model = "DG41202"
        serial = "S192837465"
        version = "3.2.1"
        ret_val = ",".join([vendor, model, serial, version]) + "\n"
        with open_close(self.instr):
            self._transport_mock.read_until.return_value = ret_val.encode("ascii")
            idn = self.instr.get_idn()

        self.assertEqual(vendor, idn.vendor)
        self.assertEqual(model, idn.model)
        self.assertEqual(serial, idn.serial)
        self.assertEqual(version, idn.version)

    def test_get_idn_excepts(self):
        """Test get_idn call"""
        vendor = "Rigol"
        model = "DG41202"
        serial = "S192837465"
        ret_val = ",".join([vendor, model, serial]) + "\n"
        with open_close(self.instr), self.assertRaises(QMI_InstrumentException):
            self._transport_mock.read_until.return_value = ret_val.encode("ascii")
            _ = self.instr.get_idn()

    def test_get_output_state(self):
        """Test instrument state query."""
        with open_close(self.instr):
            self._transport_mock.read_until.side_effect = [b"1\n", b"\n"]
            state_1 = self.instr.get_output_state()

        with open_close(self.instr):
            self._transport_mock.read_until.side_effect = [b"0\n", b"\n"]
            state_0 = self.instr.get_output_state()

        self.assertTrue(state_1)
        self.assertFalse(state_0)

    def test_get_output_state_raises_exception(self):
        """Test instrument state query raises exception at non-bool response"""
        with open_close(self.instr), self.assertRaises(QMI_InstrumentException):
            self._transport_mock.read_until.side_effect = [b"true\n"]
            self.instr.get_output_state()

    def test_set_output_state(self):
        """Test setting clock to 'internal' mode."""
        expected_call = [call(b":OUTP1 ON\n")]
        with open_close(self.instr):
            self._transport_mock.read_until.side_effect = [b'0,"No error"\n']  # For _check_error
            self.instr.set_output_state(True)
            self._transport_mock.write.assert_called_with(b"SYSTem:ERRor?\n")
            self._transport_mock.write.assert_has_calls(expected_call)

    def test_get_waveform(self):
        """Test setting a waveform."""
        waveform = b"SINusoid\n"
        expected_call = b":SOURce1:FUNCtion?\n"
        with open_close(self.instr):
            self._transport_mock.read_until.side_effect = [waveform, b'0,"No error"\n']
            response = self.instr.get_waveform()
            self._transport_mock.write.assert_called_with(expected_call)
            self.assertEqual(response, waveform.decode().strip())

    def test_set_waveform(self):
        """Test setting a waveform."""
        waveform = "SINusoid"
        expected_call = [call(f":SOURce1:FUNCtion {waveform}\n".encode("ascii"))]
        with open_close(self.instr):
            self._transport_mock.read_until.side_effect = [b'0,"No error"\n']
            self.instr.set_waveform(waveform)
            self._transport_mock.write.assert_called_with(b"SYSTem:ERRor?\n")
            self._transport_mock.write.assert_has_calls(expected_call)

    def test_get_frequency(self):
        """Test getting a frequency."""
        frequency = b"100000.0\n"
        expected_call = b":SOURce1:FREQ?\n"
        with open_close(self.instr):
            self._transport_mock.read_until.side_effect = [frequency]
            response = self.instr.get_frequency()
            self._transport_mock.write.assert_called_with(expected_call)
            self.assertEqual(response, float(frequency.decode().strip()))

    def test_get_frequency_excepts(self):
        """Test getting a frequency excepts at non-float(ish) response."""
        frequency = b"100000Hz\n"
        with open_close(self.instr), self.assertRaises(QMI_InstrumentException):
            self._transport_mock.read_until.side_effect = [frequency]
            self.instr.get_frequency()

    def test_set_frequency(self):
        """Test setting a frequency."""
        frequency = "100000.0"
        expected_call = [call(f":SOURce1:FREQ {frequency}\n".encode("ascii"))]
        with open_close(self.instr):
            self._transport_mock.read_until.side_effect = [b'0,"No error"\n']
            self.instr.set_frequency(float(frequency))
            self._transport_mock.write.assert_called_with(b"SYSTem:ERRor?\n")
            self._transport_mock.write.assert_has_calls(expected_call)

    def test_get_amplitude(self):
        """Test setting an amplitude."""
        amplitude = b"100.0\n"
        expected_call = b":SOURce1:VOLT?\n"
        with open_close(self.instr):
            self._transport_mock.read_until.side_effect = [amplitude]
            response = self.instr.get_amplitude()
            self._transport_mock.write.assert_called_with(expected_call)
            self.assertEqual(response, float(amplitude.decode().strip()))

    def test_set_amplitude(self):
        """Test setting an amplitude."""
        amplitude = "100.0"
        expected_call = [call(f":SOURce1:VOLT {amplitude}\n".encode("ascii"))]
        with open_close(self.instr):
            self._transport_mock.read_until.side_effect = [b'0,"No error"\n']
            self.instr.set_amplitude(float(amplitude))
            self._transport_mock.write.assert_called_with(b"SYSTem:ERRor?\n")
            self._transport_mock.write.assert_has_calls(expected_call)

    def test_get_offset(self):
        """Test setting an offset."""
        offset = b"-10.0\n"
        expected_call = b":SOURce1:VOLTage:OFFSet?\n"
        with open_close(self.instr):
            self._transport_mock.read_until.side_effect = [offset]
            response = self.instr.get_offset()
            self._transport_mock.write.assert_called_with(expected_call)
            self.assertEqual(response, float(offset.decode().strip()))

    def test_set_offset(self):
        """Test setting an offset."""
        offset = "-10.0"
        expected_call = [call(f":SOURce1:VOLTage:OFFSet {offset}\n".encode("ascii"))]
        with open_close(self.instr):
            self._transport_mock.read_until.side_effect = [b'0,"No error"\n']
            self.instr.set_offset(float(offset))
            self._transport_mock.write.assert_called_with(b"SYSTem:ERRor?\n")
            self._transport_mock.write.assert_has_calls(expected_call)

    def test_get_phase(self):
        """Test setting an phase."""
        phase = b"-180.0\n"
        expected_call = b":SOURce1:PHASe?\n"
        with open_close(self.instr):
            self._transport_mock.read_until.side_effect = [phase]
            response = self.instr.get_phase()
            self._transport_mock.write.assert_called_with(expected_call)
            self.assertEqual(response, float(phase.decode().strip()))

    def test_set_phase(self):
        """Test setting an phase."""
        phase = "-180.0"
        expected_call = [call(f":SOURce1:PHASe {phase}\n".encode("ascii"))]
        with open_close(self.instr):
            self._transport_mock.read_until.side_effect = [b'0,"No error"\n']
            self.instr.set_phase(float(phase))
            self._transport_mock.write.assert_called_with(b"SYSTem:ERRor?\n")
            self._transport_mock.write.assert_has_calls(expected_call)


if __name__ == '__main__':
    unittest.main()



"""Implementation of SCPI protocol primitives."""

from typing import Optional

from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.transport import QMI_Transport


class ScpiProtocol:
    """Implement SCPI protocol primitives.

    An `ScpiProtocol` instance may be instantiated by `QMI_Instrument` classes
    that talk *SCPI* over some transport.
    After instantiation, the `ScpiProtocol` primitives should be used to talk to the instrument.

    Note that the `ScpiProtocol` does not take ownership of the transport. The transport's owner is
    responsible for closing the transport even when the `ScpiProtocol` is used.

    Example::

        self._transport = transport_descriptor.create()
        self._scpi = ScpiProtocol(self._transport)

        idn = self._scpi.ask("*IDN?")
    """

    def __init__(self,
                 transport: QMI_Transport,
                 command_terminator: str = "\n",
                 response_terminator: str = "\n",
                 default_timeout: Optional[float] = None
                 ):
        """Initialize the SCPI protocol handler.

        Parameters:
            transport: Instance of `QMI_Transport` to use for sending SCPI commands to the instrument.
            command_terminator: Termination string to append when sending SCPI commands.
                This defaults to *newline* as specified by the SCPI standard.
            response_terminator: Termination string expected at the end of SCPI response messages.
                This defaults to *newline* as specified by the SCPI standard.
            default_timeout: Optional default response timeout in seconds.
                The default is to wait indefinitely until a response is received.
        """

        # The SCPI standard prescribes that a single newline character should work both for sending and receiving.
        # Just to be sure, we allow them to be overridden in case we encounter an instrument that doesn't follow
        # the standard properly.

        encoded_command_terminator  = command_terminator.encode("ascii")
        encoded_response_terminator = response_terminator.encode("ascii")

        self._transport           = transport
        self._command_terminator  = encoded_command_terminator
        self._response_terminator = encoded_response_terminator
        self._default_timeout     = default_timeout

    def write(self, cmd: str) -> None:
        """Send an SCPI command."""
        binary_cmd = cmd.encode("ascii") + self._command_terminator
        self._transport.write(binary_cmd)

    def write_raw(self, cmd: bytes) -> None:
        """Send an SCPI command already encoded as bytes"""
        binary_cmd = cmd + self._command_terminator
        self._transport.write(binary_cmd)

    def ask(self, cmd: str, timeout: Optional[float] = None, discard: bool = False) -> str:
        """Send an SCPI command, then read and return the response.

        Parameters:
            cmd: SCPI command string.
            timeout: Optional response timeout in seconds.
            discard: Discard contents in read buffer before asking. Default is False.

        Returns:
            Response message with message terminator removed.
        """

        if timeout is None:
            timeout = self._default_timeout

        if discard:
            self._transport.discard_read()

        # Send command.
        self.write(cmd)

        # Read response.
        response = self._transport.read_until(message_terminator=self._response_terminator, timeout=timeout)
        if not response.endswith(self._response_terminator):
            raise QMI_InstrumentException("Bad response")

        response = response[:-len(self._response_terminator)]

        decoded_response = response.decode("ascii")

        return decoded_response

    def read_binary_data(self,
                         read_terminator_flag: bool = True,
                         timeout: Optional[float] = None
                         ) -> bytes:
        """Read a binary data block formatted as in SCPI *definite length arbitrary block response data*.

        Parameters:
            read_terminator_flag: True to expect a message terminator after the binary data.
                This should be True (the default) if the binary data block is the last
                or only data element in the response message.
            timeout: Optional timeout in seconds.
        """

        if timeout is None:
            timeout = self._default_timeout

        # First byte: "#"
        # Second byte: decimal digit representing number of decimal digits in the length.
        header = self._transport.read(2, timeout=timeout)

        if header[0] != ord("#"):
            raise QMI_InstrumentException("Invalid binary data format, expecting '#' but got {!r}".format(header[0:1]))

        if not header[1:].isdigit():
            msg = "Invalid binary data format, expecting digit but got {!r}".format(header[1:])
            raise QMI_InstrumentException(msg)

        # Read data size.
        num_digits = int(header[1:])

        if num_digits == 0:
            raise QMI_InstrumentException("Invalid binary data format {!r}".format(header))

        header2 = self._transport.read(num_digits, timeout=timeout)
        if not header2.isdigit():
            msg = "Invalid binary data format, expecting data size but got {!r}" .format(header2)
            raise QMI_InstrumentException(msg)
        num_bytes = int(header2)

        # Read binary data.
        data = self._transport.read(num_bytes, timeout=timeout)

        if read_terminator_flag:
            # Read response terminator.
            tail = self._transport.read(len(self._response_terminator), timeout=timeout)
            if tail != self._response_terminator:
                msg = "Invalid binary data format, expecting newline but got {!r}".format(tail)
                raise QMI_InstrumentException(msg)

        return data



"""
Raspberry Pi, computer modules.

The qmi.instruments.raspberry package provides support for:
- GPIO header of various computer models.
"""
# Alternative, QMI naming convention approved names
from qmi.instruments.raspberry.gpio import RaspberryPiGPIO as RaspberryPi_Gpio



"""
Basic input/output access to the Newport 843-R power meter via USB.

This Python 3 module provides access to ASCII command interface of
the Newport 843-R-USB via its proprietary USB protocol.

This module uses PyUSB and works only on Linux.

Access permission
-----------------

To access the instrument, read/write permission to the USB device node
(/dev/usb/<bus>/<addr>) is required. This is typically accomplished
on Linux by adding a suitable rule in /etc/udev/rules.d. For example:

  /etc/udev/rules.d/99-newport_usb.rules:
  SUBYSTEM=="usb", ATTRS{idVendor}=="0bd3", ATTRS{idProduct}=="e345", GROUP="dialout", MODE="0660"

Command language
----------------

The USB access provided by this module enables the use of the ASCII command
set of the 843-R instrument as documented by Newport.
See "PMManager-User-Commands.pdf" or "Newport User Commands.pdf".

Command strings start with "$", followed by a two-letter command,
optionally followed by parameters, and ending with "\r\n".

In case of success, the instrument sends a response string starting with "*",
followed by a command-specific response format, and ending with "\n".

If an error occurs, the instrument sends a response string starting with "?",
followed by an error message and ending with "\n".

Streaming mode
--------------

In addition to the command/response flow, the instrument also supports
a streaming mode in which it spontaneously sends several power readings
per second via the USB interrupt endpoint 0x82. A sequence of undocumented
commands is required to enable this streaming mode. The streaming mode
is currently not supported by this Python module.

Protocol notes
--------------

The instrument has USB vendor ID 0x0BD3, vendor string "Freescale", product ID 0xE345, product string "843-R".

A serial number string is present in the USB descriptor.

The USB device class is 255 (vendor-specified), subclass 0.
The device supports 1 USB configuration, containing 1 interface.
The interface class is 255 (vendor-specified), subclass 255.

The interface contains 4 USB endpoints:
  - EP 0x81: interrupt IN, max 512 bytes (purpose unknown)
  - EP 0x82: interrupt IN, max 512 bytes (used in streaming mode)
  - EP 0x83: bulk IN, max 512 bytes (purpose unknown)
  - EP 0x04: bulk OUT, max 512 bytes (purpose unknown)

The request/response flow based on ASCII strings is implemented
via control transfers on the default control endpoint (EP 0x00).

To send a command string to the instrument, the computer initiates
the following control transfer:

  bmRequestType = 0x40 (host-to-device, vendor-specified, device request)
  bRequest = 2
  wValue = 0
  wIndex = 0
  data = command string (starting with "$", ending with "\\r\\n")

To read a response string from the instrument, the computer initiates
the following control transfer:

  bmRequestType = 0xc0 (device-to-host, vendor-specifid, device request)
  bRequest = 4
  wValue = 0
  wIndex = 0
  wLength = 2000

The device then responds with an answer string (starting with "*" or "?"
and ending with "\n").

Write command and read response transactions are sent in strict alternation.
Each command triggers exactly 1 response from the device.

The device-side implementation of this protocol is quite fragile.
Attempting to read a response when the instrument is not ready to
send a response will easily crash the USB interface to the point
where it can no longer respond to USB descriptor requests.
"""

import logging
from typing import List, Optional

import usb
import usb.core
import usb.util


class Newport_843R_libusb:
    """Basic input/output access to the Newport 843-R power meter via USB."""

    # USB vendor ID and product ID of Newport 843-R-USB.
    VENDOR_ID = 0x0BD3
    PRODUCT_ID = 0xE345

    # Interface index.
    INTERFACE = 0

    @classmethod
    def list_instruments(cls) -> List[str]:
        """Return serial numbers of attached Newport 843-R instruments.

        Scan the USB bus for attached Newport 843-R instruments and
        return a list of their serial numbers.
        """

        logger = logging.getLogger(__name__)
        logger.debug("Looking for attached instruments")

        instruments = []

        devs = usb.core.find(find_all=True,
                             idVendor=cls.VENDOR_ID,
                             idProduct=cls.PRODUCT_ID)
        for dev in devs:
            try:
                snr = dev.serial_number
            except Exception:
                # This happens if the user does not have permission
                # to open the USB device node /dev/usb/<bus>/<addr>.
                logger.exception("Can not read USB serial_number")
                continue
            instruments.append(snr)

        return instruments

    def __init__(self, serial_number: Optional[str]) -> None:
        """Open a connection to an attached Newport 843-R instrument.

        Note that read/write permission to the USB device node is
        required to access the instrument. This is typically accomplished
        on Linux by adding a suitable rule in /etc/udev/rules.d.

        :param serial_number: USB serial number of the device to open.
            When serial_number is None and only one device is atached,
            the single device will be selected automatically.
        :raises LookupError: When the specified instrument is not found;
            or when communication with the instrument fails.
        :raises USBError: If an error occurs at the USB level.
        """

        logger = logging.getLogger(__name__)
        logger.debug("Looking for attached instrument, serial_number=%r", serial_number)

        # Find matching instruments.
        instruments = []
        devs = usb.core.find(find_all=True,
                             idVendor=self.VENDOR_ID,
                             idProduct=self.PRODUCT_ID)
        for dev in devs:
            if serial_number is None:
                snr = None
            else:
                try:
                    snr = dev.serial_number
                except Exception:
                    # This happens if the user does not have permission
                    # to open the USB device node /dev/usb/<bus>/<addr>.
                    logger.exception("Can not read USB serial_number")
                    snr = None
            if serial_number == snr:
                # Found matching device.
                instruments.append(dev)

        # Fail if no matching instruments.
        if len(instruments) == 0:
            errmsg = "No Newport-843R instrument found"
            if serial_number is not None:
                errmsg += " with serial_number {}".format(serial_number)
            raise LookupError(errmsg)

        # Fail if multiple matching instruments.
        if len(instruments) != 1:
            raise LookupError("Multiple Newport-843R instruments found")

        # Select the single matching instrument.
        (self._dev,) = instruments

        # Detach kernel driver.
        driver_active = self._dev.is_kernel_driver_active(self.INTERFACE)
        if driver_active:
            self._dev.detach_kernel_driver(self.INTERFACE)

        # Select configuration.
        # This should also resets the USB protocol somewhat.
        self._dev.set_configuration()

        # Claim the interface.
        usb.util.claim_interface(self._dev, self.INTERFACE)

    def close(self) -> None:
        """Release USB device.

        :raises USBError: If an error occurs at the USB level.
        """

        if self._dev is not None:
            usb.util.dispose_resources(self._dev)
        self._dev = None

    def send_command(self, cmd: bytes, timeout: float = 1.0) -> None:
        """Send a command to the instrument and return the response.

        See the Newport documentation for a list of supported commands
        (PMManager-User-Commands.pdf).

        :param cmd: Command string to send to the instrument.
            The command is typically an ASCII string, starting with "$"
            and ending with "\\r\\n".
        :return: Response from the instrument.
        :raises USBError: If an error occurs on the USB level.

        """

        assert len(cmd) > 0
        assert len(cmd) < 64
        assert self._dev is not None

        bmRequestType = 0x40    # vendor-specific device request OUT
        bRequest = 2            # write request ?
        wValue = 0
        wIndex = 0
        timeout = int(1000 * timeout) + 1

        self._dev.ctrl_transfer(bmRequestType,
                                bRequest,
                                wValue,
                                wIndex,
                                cmd,
                                timeout)

    def read_response(self, timeout: float = 1.0) -> bytes:
        """Read a response string from the instrument.

        Warning: The USB interface in the instrument is extremely fickle.
        Attempting to read a response when the instrument is not ready
        to send a response, may crash the USB interface and require
        power-cycling of the instrument to recover.

        :param timeout: Maximum time to wait for the response (seconds).
        :return: Response from the instrument.
            The response is typically an ASCII string ending with "\\n".
        :raises USBError: If the device does not respond or when
            an error occurs on the USB level.
        """

        assert self._dev is not None

        bmRequestType = 0xc0    # vendor-specific device request IN
        bRequest = 4            # read request ?
        wValue = 0
        wIndex = 0
        wLength = 2000
        timeout = int(1000 * timeout) + 1

        data = self._dev.ctrl_transfer(bmRequestType,
                                       bRequest,
                                       wValue,
                                       wIndex,
                                       wLength,
                                       timeout)
        return data.tobytes()



from unittest import TestCase
from unittest.mock import patch
from unittest.mock import MagicMock
from unittest.mock import call


from dataclasses import dataclass

from typing import cast

from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.transport import QMI_Transport
from qmi.instruments.stanford_research_systems.dc205 import SRS_DC205


@dataclass
class TestMeta:
    """Test meta data."""

    transport: MagicMock
    transport_str: MagicMock
    name: MagicMock
    scpi: MagicMock
    super: MagicMock
    instr: SRS_DC205


class TestDC205(TestCase):
    """Testcase for the SRS_DC205 class."""

    def setUp(self):
        mock_transport = MagicMock(spec=QMI_Transport)
        mock_name = MagicMock()
        mock_transport_str = MagicMock()
        mock_scpi = MagicMock(spec=ScpiProtocol)
        mock_super = MagicMock()

        with patch(
            "qmi.instruments.stanford_research_systems.dc205.create_transport",
            return_value=mock_transport,
        ), patch(
            "qmi.instruments.stanford_research_systems.dc205.ScpiProtocol", mock_scpi
        ):
            instr = SRS_DC205(MagicMock(), mock_name, mock_transport_str)

        self._patcher_super = patch(
            "qmi.instruments.stanford_research_systems.dc205.super", mock_super
        )
        self._patcher_super.start()

        self._meta = TestMeta(
            transport=mock_transport,
            instr=cast(SRS_DC205, instr),
            transport_str=mock_transport_str,
            name=mock_name,
            scpi=mock_scpi(mock_transport),
            super=mock_super,
        )

    def tearDown(self):
        self._meta = None
        self._patcher_super.stop()

    def test_init(self):
        """SRS_DC205.__init__(), happy flow."""
        self.assertEqual(self._meta.instr._transport, self._meta.transport)
        self.assertEqual(self._meta.instr._scpi_protocol, self._meta.scpi)

    def test_open(self):
        """SRS_DC205.open(), happy flow."""
        self._meta.instr.open()
        self._meta.transport.open.assert_called_once_with()
        self._meta.transport.write.assert_called_once_with(b"\nTERM LF\n")
        self._meta.super().open.assert_called_once_with()

    def test_open_exception(self):
        """SRS_DC205.open(), exception on write."""
        self._meta.transport.write = MagicMock(side_effect=Exception)
        with self.assertRaises(Exception):
            self._meta.instr.open()
        self._meta.transport.close.assert_called_once_with()

    def test_close(self):
        """SRS_DC205.close(), happy flow."""
        self._meta.instr.close()
        self._meta.transport.close.assert_called_once_with()
        self._meta.super().close.assert_called_once_with()

    def test_check_error(self):
        """SRS_DC205._check_error(), happy flow."""
        self._meta.scpi.ask = MagicMock(return_value="0;0")
        self._meta.instr._check_error(MagicMock())
        self._meta.scpi.ask.assert_called_once_with("LEXE?; LCME?")

    def test_check_error_execution_error(self):
        """SRS_DC205._check_error(), execution error handling."""
        self._meta.scpi.ask = MagicMock(return_value="1;0")
        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr._check_error(MagicMock())
        self._meta.scpi.ask.assert_called_once_with("LEXE?; LCME?")

    def test_check_error_command_error(self):
        """SRS_DC205._check_error(), command error handling."""
        self._meta.scpi.ask = MagicMock(return_value="0;1")
        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr._check_error(MagicMock())
        self._meta.scpi.ask.assert_called_once_with("LEXE?; LCME?")

    def test_check_error_invalid_response(self):
        """SRS_DC205._check_error(), invalid response handling."""
        self._meta.scpi.ask = MagicMock(return_value="")
        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr._check_error(MagicMock())
        self._meta.scpi.ask.assert_called_once_with("LEXE?; LCME?")

    def test_set_command(self):
        """SRS_DC205._set_command(), happy flow."""
        mock_cmd = MagicMock()
        self._meta.instr._check_is_open = MagicMock()
        self._meta.instr._check_error = MagicMock()
        self._meta.instr._set_command(mock_cmd)
        self._meta.scpi.write.assert_called_once_with(mock_cmd)
        self._meta.instr._check_is_open.assert_called_once_with()
        self._meta.instr._check_error.assert_called_once_with(mock_cmd)

    def test_ask_float(self):
        """SRS_DC205._ask_float(), happy flow."""
        mock_cmd = MagicMock()
        mock_resp = "0.01"
        self._meta.instr._check_is_open = MagicMock()
        self._meta.scpi.ask = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr._ask_float(mock_cmd)
        self._meta.scpi.ask.assert_called_once_with(mock_cmd)
        self._meta.instr._check_is_open.assert_called_once_with()
        self.assertEqual(rt_val, float(mock_resp))

    def test_ask_float_value_error(self):
        """SRS_DC205._ask_float(), value error handling."""
        mock_cmd = MagicMock()
        mock_resp = "notafloat"
        self._meta.instr._check_is_open = MagicMock()
        self._meta.scpi.ask = MagicMock(return_value=mock_resp)
        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr._ask_float(mock_cmd)
        self._meta.scpi.ask.assert_called_once_with(mock_cmd)
        self._meta.instr._check_is_open.assert_called_once_with()

    def test_ask_token(self):
        """SRS_DC205._ask_token(), happy flow, token."""
        mock_cmd = MagicMock()
        mock_tokens = [MagicMock(), MagicMock()]
        mock_resp = "1"
        self._meta.instr._check_is_open = MagicMock()
        self._meta.scpi.ask = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr._ask_token(mock_cmd, mock_tokens)
        self._meta.instr._check_is_open.assert_called_once_with()
        self._meta.scpi.ask.assert_called_once_with(mock_cmd)
        self.assertEqual(rt_val, int(mock_resp))

    def test_ask_token_in_tokens(self):
        """SRS_DC205._ask_token(), happy flow, index of token."""
        mock_cmd = MagicMock()
        mock_resp = "1"
        mock_index = 3
        mock_tokens = [MagicMock()] * (mock_index * 2)
        mock_tokens[mock_index] = mock_resp
        self._meta.instr._check_is_open = MagicMock()
        self._meta.scpi.ask = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr._ask_token(mock_cmd, mock_tokens)
        self._meta.instr._check_is_open.assert_called_once_with()
        self._meta.scpi.ask.assert_called_once_with(mock_cmd)
        self.assertEqual(
            rt_val, mock_index
        )  # rt_val is the index of the token in mock_tokens

    def test_ask_token_no_int(self):
        """SRS_DC205._ask_token(), response is not an int handling."""
        mock_cmd = MagicMock()
        mock_tokens = [MagicMock(), MagicMock()]
        mock_resp = "notaint"
        self._meta.instr._check_is_open = MagicMock()
        self._meta.scpi.ask = MagicMock(return_value=mock_resp)
        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr._ask_token(mock_cmd, mock_tokens)
        self._meta.instr._check_is_open.assert_called_once_with()
        self._meta.scpi.ask.assert_called_once_with(mock_cmd)

    def test_ask_token_invalid_response(self):
        """SRS_DC205._ask_token(), invalid response handling."""
        mock_cmd = MagicMock()
        mock_tokens = []
        mock_resp = "5"
        self._meta.instr._check_is_open = MagicMock()
        self._meta.scpi.ask = MagicMock(return_value=mock_resp)
        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr._ask_token(mock_cmd, mock_tokens)
        self._meta.instr._check_is_open.assert_called_once_with()
        self._meta.scpi.ask.assert_called_once_with(mock_cmd)

    def test_reset(self):
        """SRS_DC205.reset(), happy flow."""
        self._meta.instr._check_is_open = MagicMock()
        self._meta.instr._check_error = MagicMock()
        self._meta.instr.reset()
        self._meta.scpi.ask.assert_has_calls(
            [
                call("LEXE?; LCME?", timeout=self._meta.instr.RESET_RESPONSE_TIMEOUT),
                call("*OPC?", timeout=self._meta.instr.RESET_RESPONSE_TIMEOUT),
            ]
        )
        self._meta.scpi.write.assert_called_once_with("*RST")
        self._meta.instr._check_is_open.assert_called_once_with()
        self._meta.instr._check_error.assert_called_once_with("*RST")

    def test_get_idn(self):
        """SRS_DC205.get_idn(), happy flow."""
        self._meta.instr._check_is_open = MagicMock()
        self._meta.scpi.ask = MagicMock(return_value="VENDOR,MODEL,SERIAL,VERSION")
        rt_val = self._meta.instr.get_idn()
        self._meta.instr._check_is_open.assert_called_once_with()
        self._meta.scpi.ask.assert_called_once_with("*IDN?")
        self.assertEqual(rt_val.vendor, "VENDOR")
        self.assertEqual(rt_val.model, "MODEL")
        self.assertEqual(rt_val.serial, "SERIAL")
        self.assertEqual(rt_val.version, "VERSION")

    def test_get_idn_invalid_response(self):
        """SRS_DC205.get_idn(), invalid response handling."""
        self._meta.instr._check_is_open = MagicMock()
        self._meta.scpi.ask = MagicMock(return_value="")
        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr.get_idn()
        self._meta.instr._check_is_open.assert_called_once_with()
        self._meta.scpi.ask.assert_called_once_with("*IDN?")

    def test_get_range(self):
        """SRS_DC205.get_range(), happy flow."""
        map_resp = [
            {"index": 0, "value": 1},
            {"index": 1, "value": 10},
            {"index": 2, "value": 100},
        ]
        for item in map_resp:
            self._meta.instr._ask_token = MagicMock(return_value=item["index"])
            rt_val = self._meta.instr.get_range()
            self._meta.instr._ask_token.assert_called_once_with(
                cmd="RNGE?", tokens=["RANGE1", "RANGE10", "RANGE100"]
            )
            self.assertEqual(rt_val, item["value"])

    def test_set_range(self):
        """SRS_DC205.set_range(), happy flow."""
        mock_volt_range = [1, 10, 100]
        for item in mock_volt_range:
            self._meta.instr._set_command = MagicMock()
            self._meta.instr.set_range(item)
            self._meta.instr._set_command.assert_called_once_with(f"RNGE RANGE{item}")

    def test_set_range_invalid_range(self):
        """SRS_DC205.set_range(), invalid range handling."""
        mock_volt_range = MagicMock()
        self._meta.instr._set_command = MagicMock()
        with self.assertRaises(ValueError):
            self._meta.instr.set_range(mock_volt_range)

    def test_get_output_enabled(self):
        """SRS_DC205.get_output_enabled(), happy flow."""
        mock_resp = [0, 1]
        for item in mock_resp:
            self._meta.instr._ask_token = MagicMock(return_value=item)
            rt_val = self._meta.instr.get_output_enabled()
            self._meta.instr._ask_token.assert_called_once_with(
                cmd="SOUT?", tokens=["OFF", "ON"]
            )
            self.assertEqual(rt_val, bool(item))

    def test_set_output_enabled(self):
        """SRS_DC205.set_output_enabled(), happy flow."""
        mock_enable = True
        self._meta.instr._set_command = MagicMock()
        self._meta.instr.set_output_enabled(mock_enable)
        self._meta.instr._set_command.assert_called_once_with(
            "SOUT {}".format(int(mock_enable))
        )

    def test_get_voltage(self):
        """SRS_DC205.get_voltage(), happy flow."""
        mock_resp = MagicMock()
        self._meta.instr._ask_float = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr.get_voltage()
        self._meta.instr._ask_float.assert_called_once_with("VOLT?")
        self.assertEqual(rt_val, mock_resp)

    def test_set_voltage(self):
        """SRS_DC205.set_voltage(), happy flow."""
        mock_voltage = 0.01
        self._meta.instr._set_command = MagicMock()
        self._meta.instr.set_voltage(mock_voltage)
        self._meta.instr._set_command.assert_called_once_with(
            "VOLT {:.6f}".format(mock_voltage)
        )

    def test_get_output_floating(self):
        """SRS_DC205.get_output_floating(), happy flow."""
        mock_resp = [0, 1]
        for item in mock_resp:
            self._meta.instr._ask_token = MagicMock(return_value=item)
            rt_val = self._meta.instr.get_output_floating()
            self._meta.instr._ask_token.assert_called_once_with(
                cmd="ISOL?", tokens=["GROUND", "FLOAT"]
            )
            self.assertEqual(rt_val, bool(item))

    def test_set_output_floating(self):
        """SRS_DC205.set_output_floating(), happy flow."""
        mock_enable = True
        self._meta.instr._set_command = MagicMock()
        self._meta.instr.set_output_floating(mock_enable)
        self._meta.instr._set_command.assert_called_once_with(
            "ISOL {}".format(int(mock_enable))
        )

    def test_get_sensing_enabled(self):
        """SRS_DC205.get_sensing_enabled(), happy flow."""
        mock_resp = [0, 1]
        for item in mock_resp:
            self._meta.instr._ask_token = MagicMock(return_value=item)
            rt_val = self._meta.instr.get_sensing_enabled()
            self._meta.instr._ask_token.assert_called_once_with(
                cmd="SENS?", tokens=["TWOWIRE", "FOURWIRE"]
            )
            self.assertEqual(rt_val, bool(item))

    def test_set_sensing_enabled(self):
        """SRS_DC205.set_sensing_enabled(), happy flow."""
        mock_enable = True
        self._meta.instr._set_command = MagicMock()
        self._meta.instr.set_sensing_enabled(mock_enable)
        self._meta.instr._set_command.assert_called_once_with(
            "SENS {}".format(int(mock_enable))
        )

    def test_get_interlock_status(self):
        """SRS_DC205.get_interlock_status(), happy flow."""
        mock_resp = [0, 1]
        for item in mock_resp:
            self._meta.instr._ask_token = MagicMock(return_value=item)
            rt_val = self._meta.instr.get_interlock_status()
            self._meta.instr._ask_token.assert_called_once_with(
                cmd="ILOC?", tokens=["OPEN", "CLOSED"]
            )
            self.assertEqual(rt_val, bool(item))

    def test_get_overloaded(self):
        """SRS_DC205.get_overloaded(), happy flow."""
        mock_resp = [0, 1]
        for item in mock_resp:
            self._meta.instr._ask_token = MagicMock(return_value=item)
            rt_val = self._meta.instr.get_overloaded()
            self._meta.instr._ask_token.assert_called_once_with(
                cmd="OVLD?", tokens=["OKAY", "OVLD"]
            )
            self.assertEqual(rt_val, bool(item))



import unittest
from unittest.mock import MagicMock

from qmi.core.transport import QMI_Transport
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.exceptions import QMI_InstrumentException


class TestScpiProtocol(unittest.TestCase):

    def test_constructor(self):

        transport = MagicMock()

        ScpiProtocol(transport)
    
    def test_write(self):
        # arrange
        command = "test"
        expected_call = b"test\n"

        transport_mock = MagicMock(spec=QMI_Transport)
       
        # act
        scpi_protocol = ScpiProtocol(transport_mock)
        scpi_protocol.write(command)

        # assert
        transport_mock.write.assert_called_with(expected_call)

    def test_write_carrage_return(self):
        # arrange
        command = "test"
        command_terminator = "\r"
        expected_call = b"test\r"
        transport_mock = MagicMock(spec=QMI_Transport)

        # act
        scpi_protocol = ScpiProtocol(transport_mock, command_terminator)
        scpi_protocol.write(command)
        
        # assert
        transport_mock.write.assert_called_with(expected_call)

    def test_ask(self):
        # arrange
        command = "test"
        expected_call = b"test\n"
        transport_mock = MagicMock(spec=QMI_Transport)
        transport_mock.read_until.return_value = b"intresting result\n"
       
        # act
        scpi_protocol = ScpiProtocol(transport_mock)
        response = scpi_protocol.ask(command)

        # assert
        transport_mock.write.assert_called_with(expected_call)
        transport_mock.discard_read.assert_not_called()
        transport_mock.read_until.assert_called_with(message_terminator=b'\n', timeout=None)
        self.assertEqual(response, "intresting result")

    def test_ask_discard(self):
        # arrange
        command = "test"
        expected_call = b"test\n"
        transport_mock = MagicMock(spec=QMI_Transport)
        transport_mock.read_until.return_value = b"intresting result\n"
       
        # act
        scpi_protocol = ScpiProtocol(transport_mock)
        response = scpi_protocol.ask(command, discard=True)

        # assert
        transport_mock.write.assert_called_with(expected_call)
        transport_mock.discard_read.assert_called_once()
        transport_mock.read_until.assert_called_with(message_terminator=b'\n', timeout=None)
        self.assertEqual(response, "intresting result")

    def test_ask_raises_on_bad_response(self):
        # arrange
        command = "test"
        transport_mock = MagicMock(spec=QMI_Transport)
        transport_mock.read_until.return_value = b"intresting result\r"
       
        # act
        # assert
        scpi_protocol = ScpiProtocol(transport_mock)
        with self.assertRaises(QMI_InstrumentException):
            scpi_protocol.ask(command)

    def test_read_binary_data(self):
        #arrange
        transport_mock = MagicMock(spec=QMI_Transport)
        expected_data = b"1234567890"
        transport_mock.read.side_effect = [b"#2", b"10", expected_data, b"\n"]

        #act
        scpi_protocol = ScpiProtocol(transport_mock)
        data = scpi_protocol.read_binary_data()

        #assert
        self.assertEqual(data, expected_data)

    def test_read_binary_data_raise_invalid_first_byte(self):
        #arrange
        transport_mock = MagicMock(spec=QMI_Transport)
        transport_mock.read.side_effect = [b"-2"]

        #act
        #assert
        scpi_protocol = ScpiProtocol(transport_mock)
        with self.assertRaises(QMI_InstrumentException):
            scpi_protocol.read_binary_data()
       
    def test_read_binary_data_raise_invalid_second_byte(self):
        #arrange
        transport_mock = MagicMock(spec=QMI_Transport)
        transport_mock.read.side_effect = [b"##"]

        #act
        #assert
        scpi_protocol = ScpiProtocol(transport_mock)
        with self.assertRaises(QMI_InstrumentException):
            scpi_protocol.read_binary_data()
       
    def test_read_binary_data_raise_when_second_byte_is_zero(self):
        #arrange
        transport_mock = MagicMock(spec=QMI_Transport)
        transport_mock.read.side_effect = [b"#0"]

        #act
        #assert
        scpi_protocol = ScpiProtocol(transport_mock)
        with self.assertRaises(QMI_InstrumentException):
            scpi_protocol.read_binary_data()

    def test_read_binary_data_raise_size_not_digits(self):
        #arrange
        transport_mock = MagicMock(spec=QMI_Transport)
        transport_mock.read.side_effect = [b"#2", b"!@"]
   
        #act
        #assert
        scpi_protocol = ScpiProtocol(transport_mock)
        with self.assertRaises(QMI_InstrumentException):
            scpi_protocol.read_binary_data()

    def test_read_binary_data_raise_no_valid_response_terminator(self):
        #arrange
        transport_mock = MagicMock(spec=QMI_Transport)
        expected_data = b"1234567890"
        transport_mock.read.side_effect = [b"#2", b"10", expected_data, b"\b"]

        #act
        #assert
        scpi_protocol = ScpiProtocol(transport_mock)
        with self.assertRaises(QMI_InstrumentException):
            scpi_protocol.read_binary_data()




"""Unit test for the Cobolt 06-01 series diode laser."""

import unittest
from unittest.mock import MagicMock, patch

import qmi
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.transport import QMI_SerialTransport
from qmi.instruments.cobolt import Hubner_Cobolt0601


class TestLaser_06_01(unittest.TestCase):

    def setUp(self):
        qmi.start("TestContext")
        self._transport_mock = MagicMock(spec=QMI_SerialTransport)
        with patch(
                'qmi.instruments.cobolt.laser_06_01.create_transport',
                return_value=self._transport_mock):
            self.instr = qmi.make_instrument("instr", Hubner_Cobolt0601, "transport_descriptor")

    def tearDown(self):
        qmi.stop()

    def test_open_close(self):
        self.instr.open()
        self._transport_mock.open.assert_called_once_with()
        self.instr.close()
        self._transport_mock.close.assert_called_once_with()

    def test_get_idn(self):
        self._transport_mock.read_until.return_value = b"12345\r\n"
        self.instr.open()
        idn = self.instr.get_idn()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"gsn?\r")
        self.assertEqual(idn.vendor, "Cobolt")
        self.assertEqual(idn.model, "06-01")
        self.assertEqual(idn.serial, "12345")

    def test_reset(self):
        self._transport_mock.read_until.return_value = b"OK\r\n"
        self.instr.open()
        self.instr.reset()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"@cob1\r")

    def test_reset_fail(self):
        self._transport_mock.read_until.return_value = b"haha\r\n"
        self.instr.open()
        with self.assertRaises(QMI_InstrumentException):
            self.instr.reset()
        self.instr.close()

    def test_get_laser_on_state(self):
        self._transport_mock.read_until.return_value = b"1\r\n"
        self.instr.open()
        laser_on_state = self.instr.get_laser_on_state()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"l?\r")
        self.assertTrue(laser_on_state)

    def test_get_laser_on_state_fail(self):
        self._transport_mock.read_until.return_value = b"123\r\n"
        self.instr.open()
        with self.assertRaises(QMI_InstrumentException):
            laser_on_state = self.instr.get_laser_on_state()
        self._transport_mock.read_until.return_value = b"LASER_ON\r\n"
        with self.assertRaises(QMI_InstrumentException):
            laser_on_state = self.instr.get_laser_on_state()
        self.instr.close()

    def test_set_laser_on_state(self):
        self._transport_mock.read_until.return_value = b"OK\r\n"
        self.instr.open()
        self.instr.set_laser_on_state(True)
        self._transport_mock.write.assert_called_once_with(b"l1\r")
        self._transport_mock.write.reset_mock()
        self.instr.set_laser_on_state(False)
        self._transport_mock.write.assert_called_once_with(b"l0\r")
        self.instr.close()

    def test_get_operating_hours(self):
        self._transport_mock.read_until.return_value = b"123.4\r\n"
        self.instr.open()
        operating_hours = self.instr.get_operating_hours()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"hrs?\r")
        self.assertAlmostEqual(operating_hours, 123.4)

    def test_get_interlock_state(self):
        self._transport_mock.read_until.return_value = b"0\r\n"
        self.instr.open()
        interlock = self.instr.get_interlock_state()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"ilk?\r")
        self.assertFalse(interlock)

    def test_get_output_power_setpoint(self):
        self._transport_mock.read_until.return_value = b"0.001234\r\n"
        self.instr.open()
        power = self.instr.get_output_power_setpoint()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"p?\r")
        self.assertAlmostEqual(power, 0.001234, places=8)

    def test_get_output_power_setpoint_fail(self):
        self._transport_mock.read_until.return_value = b"overflow\r\n"
        self.instr.open()
        with self.assertRaises(QMI_InstrumentException):
            power = self.instr.get_output_power_setpoint()
        self.instr.close()

    def test_set_output_power_setpoint(self):
        self._transport_mock.read_until.return_value = b"OK\r\n"
        self.instr.open()
        self.instr.set_output_power_setpoint(0.0121)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"p 0.012100\r")

    def test_get_output_power(self):
        self._transport_mock.read_until.return_value = b"0.00345\r\n"
        self.instr.open()
        power = self.instr.get_output_power()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"pa?\r")
        self.assertAlmostEqual(power, 0.00345, places=8)

    def test_get_drive_current(self):
        self._transport_mock.read_until.return_value = b"50.67\r\n"
        self.instr.open()
        current = self.instr.get_drive_current()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"i?\r")
        self.assertAlmostEqual(current, 50.67)

    def test_set_drive_current(self):
        self._transport_mock.read_until.return_value = b"OK\r\n"
        self.instr.open()
        self.instr.set_drive_current(60.1)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"slc 60.100000\r")

    def test_set_constant_power_mode(self):
        self._transport_mock.read_until.return_value = b"OK\r\n"
        self.instr.open()
        self.instr.set_constant_power_mode()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"cp\r")

    def test_set_constant_current_mode(self):
        self._transport_mock.read_until.return_value = b"OK\r\n"
        self.instr.open()
        self.instr.set_constant_current_mode()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"ci\r")

    def test_get_fault(self):
        self._transport_mock.read_until.return_value = b"0\r\n"
        self.instr.open()
        fault = self.instr.get_fault()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"f?\r")
        self.assertEqual(fault, 0)

    def test_clear_fault(self):
        self._transport_mock.read_until.return_value = b"OK\r\n"
        self.instr.open()
        self.instr.clear_fault()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"cf\r")

    def test_set_modulation_mode(self):
        self._transport_mock.read_until.return_value = b"OK\r\n"
        self.instr.open()
        self.instr.set_modulation_mode()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"em\r")

    def test_get_digital_modulation_state(self):
        self._transport_mock.read_until.return_value = b"1\r\n"
        self.instr.open()
        modulation = self.instr.get_digital_modulation_state()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"gdmes?\r")
        self.assertTrue(modulation)

    def test_set_digital_modulation_state(self):
        self._transport_mock.read_until.return_value = b"OK\r\n"
        self.instr.open()
        self.instr.set_digital_modulation_state(True)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"sdmes 1\r")

    def test_get_analog_modulation_state(self):
        self._transport_mock.read_until.return_value = b"0\r\n"
        self.instr.open()
        modulation = self.instr.get_analog_modulation_state()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"games?\r")
        self.assertFalse(modulation)

    def test_set_analog_modulation_state(self) -> None:
        self._transport_mock.read_until.return_value = b"OK\r\n"
        self.instr.open()
        self.instr.set_analog_modulation_state(False)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"sames 0\r")

    def test_get_operating_mode(self):
        self._transport_mock.read_until.return_value = b"2\r\n"
        self.instr.open()
        mode = self.instr.get_operating_mode()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"gom?\r")
        self.assertEqual(mode, 2)

    def test_get_operating_mode_fail(self):
        self._transport_mock.read_until.return_value = b"error\r\n"
        self.instr.open()
        with self.assertRaises(QMI_InstrumentException):
            mode = self.instr.get_operating_mode()
        self.instr.close()

    def test_get_analog_low_impedance_state(self):
        self._transport_mock.read_until.return_value = b"0\r\n"
        self.instr.open()
        impedance = self.instr.get_analog_low_impedance_state()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"galis?\r")
        self.assertFalse(impedance)

    def test_set_analog_low_impedance_state(self):
        self._transport_mock.read_until.return_value = b"OK\r\n"
        self.instr.open()
        self.instr.set_analog_low_impedance_state(True)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"salis 1\r")

    def test_get_autostart_state(self):
        self._transport_mock.read_until.return_value = b"1\r\n"
        self.instr.open()
        autostart = self.instr.get_autostart_state()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"@cobas?\r")
        self.assertTrue(autostart)

    def test_set_autostart_state(self):
        self._transport_mock.read_until.return_value = b"OK\r\n"
        self.instr.open()
        self.instr.set_autostart_state(False)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"@cobas 0\r")


if __name__ == '__main__':
    unittest.main()



"""
OZ Optics, attenuator and polarization controller.

The qmi.instruments.adwin package provides support for:
- DD-100-MC attenuator
- Electric Polarization Controller (EPC).
"""
from qmi.instruments.ozoptics.dd100mc import OZO_AttenuatorPosition
# Alternative, QMI naming convention approved names
from qmi.instruments.ozoptics.dd100mc import OZOptics_DD100MC as OzOptics_Dd100Mc
from qmi.instruments.ozoptics.epc_driver import OzOptics_EpcDriver as OzOptics_EPC
